/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.css;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.annotations.obfuscate.SkipLevel;
import ro.sync.annotations.obfuscate.SkipObfuscate;

/**
 * A CSS counter identified by it's name.
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PRIVATE)
@SkipObfuscate(classes = SkipLevel.NOT_SPECIFIED, fields = SkipLevel.NOT_SPECIFIED, methods = SkipLevel.PUBLIC)
public class CSSCounter {
  /**
   * Current counter value.
   */
  private int cValue = 0;
  /**
   * The counter name.
   */
  private final String cName;
  /**
   * The initial value of the counter.
   */
  private int initialValue;
  /**
   * <code>true</code> to ignore next increment. 
   * This field value can be set to <code>true</code> when the current counter 
   * is set to an element deleted with track changes.
   */
  private boolean ignoreNextIncrement = false;
  
  /**
   * Constructor.
   * 
   * @param name The name of the counter.
   */
  public CSSCounter(String name) {
    this(name, 0);
  }

  /**
   * Constructor.
   *  
   * @param name The name of the counter.
   * @param defaultValue The default value.
   */
  public CSSCounter(String name, int defaultValue) {
    this.cName = name;
    this.cValue = defaultValue;
    this.initialValue = defaultValue;  
  }
  
  /**
   * Increment the counter.
   * 
   * @param increment The increment value. 
   * @param ignoreNextIncremenet  <code>true</code> when the current counter is 
   * set to an element deleted with track changes, and the next increment should
   * be ignored.
   */
  public void increment(int increment, boolean ignoreNextIncremenet) {
    if (!ignoreNextIncrement) {
      cValue+=increment;
    }
    this.ignoreNextIncrement = ignoreNextIncremenet;
  }
  
  /**
   * Increment the counter.
   * 
   * @param increment The increment value. 
   */
  public void increment(int increment) {
    increment(increment, false);
  }
  
  /**
   * Reset the counter to the initial value.
   */
  public void reset() {
    this.ignoreNextIncrement = false;
    this.cValue = initialValue;
  }
  
  /**
   * @return The current value of the counter.
   */
  public int getValue() {
    return cValue;
  }
  
  /**
   * @return The counter name.
   */
  public String getName() {
    return cName;
  }
  
  /**
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString() {
    return cName + ", initial: " + initialValue + " cValue: " + cValue;
  }

  /**
   * @param defValue
   */
  public void setDefaultValue(int defValue) {
    this.initialValue = defValue;
    this.cValue = defValue;
  }
  
  /**
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (obj instanceof CSSCounter) {
      CSSCounter counter = (CSSCounter) obj;
      return cName.equals(counter.cName) && initialValue == counter.initialValue 
          && cValue == counter.cValue;
    }
    return false;
  }
  
  /**
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    int hash = cName.hashCode();
    hash += (cValue + "|" + initialValue).intern().hashCode();
    return hash;
  }
  
}