/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ro.sync.ecss.css;

import java.lang.reflect.Array;
import java.util.ArrayList;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.annotations.obfuscate.SkipLevel;
import ro.sync.annotations.obfuscate.SkipObfuscate;
import ro.sync.exml.view.graphics.Color;
import ro.sync.exml.view.graphics.Font;
import ro.sync.exml.view.graphics.FontMetrics;
import ro.sync.util.Equaler;
import ro.sync.util.NumberParserUtil;

/**
 * Represents the <b>computed</b> style properties for a particular element. 
 * These styles are taken from the CSS.
 * 
 * See: http://www.w3.org/TR/CSS21/cascade.html#q1
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PRIVATE)
@SkipObfuscate(classes = SkipLevel.PRIVATE, fields = SkipLevel.PRIVATE, methods = SkipLevel.PRIVATE)
public class Styles implements Cloneable {

  /**
   * This is the factor:  ex-height / em-height
   * The ex unit is defined by the font's x-height. The x-height is so 
   * called because it is often equal to the height of the lowercase "x". 
   * However, an ex is defined even for fonts that don't contain an "x".
   */
  public static final float EX_FACTOR = 0.6f;

  /**
   * Value for the "normal" in the font-weight propery.
   * Spec: normal  Same as '400'.
   */
  public static final int FONT_WEIGHT_NORMAL = 400;

  /**
   * Value for the "bold" in the font-weight propery.
   * Spec: normal  Same as '700'.
   */
  static final int FONT_WEIGHT_BOLD = 700;

  /**
   * Named value for the "border-width" property: "thin"
   */
  static final int BORDER_WIDTH_THIN = 1;

  /**
   * Named value for the "border-width" property: "medium"
   */
  static final int BORDER_WIDTH_MEDIUM = 3;

  /**
   * Named value for the "border-width" property: "thick"
   */
  static final int BORDER_WIDTH_THICK = 5;

  /**
   * Constants counter
   */
  public static int counter = 0;
  /**
   * Background color.
   */
  public static final int KEY_BACKGROUND_COLOR = counter ++;

  /**
   * Border bottom color
   */
  public static final int KEY_BORDER_BOTTOM_COLOR = counter ++;

  /**
   * Border bottom style
   */
  public static final int KEY_BORDER_BOTTOM_STYLE = counter ++;

  /**
   * Border bottom width
   */
  public static final int KEY_BORDER_BOTTOM_WIDTH = counter ++;

  /**
   * Border left color
   */
  public static final int KEY_BORDER_LEFT_COLOR = counter ++;

  /**
   * Border right style
   */
  public static final int KEY_BORDER_LEFT_STYLE = counter ++;

  /**
   * Border left width
   */
  public static final int KEY_BORDER_LEFT_WIDTH = counter ++;

  /**
   * Border right color
   */
  public static final int KEY_BORDER_RIGHT_COLOR = counter ++;

  /**
   * Border right style
   */
  public static final int KEY_BORDER_RIGHT_STYLE = counter ++;

  /**
   * Border right width
   */
  public static final int KEY_BORDER_RIGHT_WIDTH = counter ++;

  /**
   * Border top color
   */
  public static final int KEY_BORDER_TOP_COLOR = counter ++;

  /**
   * Border top style
   */
  public static final int KEY_BORDER_TOP_STYLE = counter ++;

  /**
   * Border top width
   */
  public static final int KEY_BORDER_TOP_WIDTH = counter ++;

  /**
   * Foreground
   */
  public static final int KEY_FOREGROUND_COLOR = counter ++;

  /**
   * Generated mixed Content
   */
  public static final int KEY_MIXED_CONTENT = counter ++;

  /**
   * Display type
   */
  public static final int KEY_DISPLAY = counter ++;

  /**
   * Used font
   */
  public static final int KEY_FONT = counter ++;

  /**
   * Font weight
   */
  public static final int KEY_FONT_WEIGHT = counter ++;

  /**
   * Line height multiplier
   */
  public static final int KEY_LINE_HEIGHT_MULTIPLIER = counter ++;

  /**
   * List style type
   */
  public static final int KEY_LIST_STYLE_TYPE = counter ++;

  /**
   * Margin dimensions
   */
  public static final int KEY_MARGIN_BOTTOM = counter ++;

  /**
   * Margin left
   */
  public static final int KEY_MARGIN_LEFT = counter ++;

  /**
   * Margin right
   */
  public static final int KEY_MARGIN_RIGHT = counter ++;

  /**
   * Margin top
   */
  public static final int KEY_MARGIN_TOP = counter ++;

  /**
   * Padding dimensions
   */
  public static final int KEY_PADDING_BOTTOM = counter ++;
  
  /**
   * Pad left
   */
  public static final int KEY_PADDING_LEFT = counter ++;

  /**
   * Pad right
   */
  public static final int KEY_PADDING_RIGHT = counter ++;

  /**
   * Pad top
   */
  public static final int KEY_PADDING_TOP = counter ++;

  /**
   * Width. Applies to blocks. It can be null.
   */
  public static final int KEY_WIDTH = counter ++;
  
  /**
   * Height. Applies to images for now. It can be null.
   */
  public static final int KEY_HEIGHT = counter ++;

  /**
   * Minimum width. Applies to blocks. It can be null.
   */
  public static final int KEY_MIN_WIDTH = counter ++;

  /**
   * Maximum width. Applies to blocks. It can be null.
   */
  public static final int KEY_MAX_WIDTH = counter ++;

  /**
   * Text align
   */
  public static final int KEY_TEXT_ALIGN = counter ++;

  /**
   * Whitespace
   */
  public static final int KEY_WHITESPACE = counter ++;

  /**
   * True if this view is editable
   */
  public static final int KEY_EDITABLE = counter ++;

  /**
   * Key for the border spacing array
   */
  public static final int KEY_BORDER_SPACING = counter ++;

  /**
   * Key for the 'visibility' property.
   */
  public static final int KEY_VISIBITY = counter ++;

  /**
   * Key used to store the property 'empty-cells' value.
   */
  public static final int KEY_EMPTY_CELLS = counter ++;

  /**
   * Key used to store the property 'foldable' value.
   */
  public static final int KEY_FOLDABLE = counter ++;

  /**
   * Key used to store the property 'non foldable child name' value.
   */
  public static final int KEY_NON_FOLDABLE_CHILD_NAME = counter ++;

  /**
   * Key for vertical align.
   */
  public static final int KEY_VERTICAL_ALIGN = counter ++;
  
  /**
   * The key for counter-reset .
   */
  public static final int KEY_COUNTER_RESET = counter ++;

  /**
   * The key for counter-increment .
   */
  public static final int KEY_COUNTER_INCREMENT = counter ++;
  /**
   * Key used to override the display property defined in the css.
   */
  public static final int KEY_IMPOSED_DISPLAY = counter ++;

  /**
   * Key used to store 'text-decoration' property.
   */
  public static final int KEY_TEXT_DECORATION = counter ++;

  /**
   * Key used to store color for text decoration.
   */
  public static final int KEY_TEXT_DECORATION_COLOR = counter ++;

  /**
   * Key used to store the URL for link elements.
   */
  public static final int KEY_LINK_URL = counter ++;
  /**
   * Key used to hide sentinel markers.
   */
  public static final int KEY_DISPLAY_TAGS = counter ++;
  
  /**
   * Whitespace specified directly on the node
   */
  public static final int KEY_DIRECT_WHITESPACE = counter ++;
  /**
   * Key set for filtered out nodes.
   */
  public static final int KEY_FILTERED_OUT = counter ++;
  
  /**
   * Text transform
   */
  public static final int KEY_TEXT_TRANSFORM = counter ++;
  
  /**
   * Key used to store the property 'show-placeholder' value.
   */
  public static final int KEY_SHOW_PLACEHOLDER = counter ++;
  
  /**
   * Key used to store the property 'placeholder-content' value.
   */
  public static final int KEY_PLACEHOLDER_CONTENT = counter ++;
  
  /**
   * Key used to store the property 'folded' value.
   */
  public static final int KEY_FOLDED = counter ++;
  
  /**
   * Key used to store the property '-oxy-tags-background-color' value.
   */
  public static final int KEY_TAGS_BACKGROUND_COLOR = counter ++;
  
  /**
   * Key used to store the property '-oxy-tags-color' value.
   */
  public static final int KEY_TAGS_COLOR = counter ++;
  
  /**
   * Key used to store the property 'background-image' value.
   */
  public static final int KEY_BACKGROUND_IMAGE = counter ++;
  
  /**
   * Key used to store the property 'background-repeat' value.
   */
  public static final int KEY_BACKGROUND_REPEAT = counter ++;
  
  /**
   * Key used to store the property 'background-position' value.
   */
  public static final int KEY_BACKGROUND_POSITION = counter ++; 


  /**
   * The styles map
   */
  private Object[] stylesArray = new Object[counter];

  private boolean canBeCached;

  /**
   * Empty string array constant.
   */
  private static final String[] EMPTY_STRING_ARRAY = new String[0];

  /**
   * Empty content
   */
  private static final StaticContent[] EMPTY_CONTENT = new StaticContent[0];

  /**
   * Empty counters declaration.
   */
  private static final CSSCounter[] EMPTY_COUNTERS = new CSSCounter[0];

  /**
   * Empty counters increment array.
   */
  private static final CSSCounterIncrement[] EMPTY_COUNTERS_INCREMENT = new CSSCounterIncrement[0];

  /**
   * Default border spacing
   */
  private static final int[] EMPTY_BORDER_SPACING = new int[] { 0, 0 };

  /**
   * Zero relative length
   */
  private static final RelativeLength ZERO_RELATIVE_LENGTH = RelativeLength.createAbsolute(0);

  /**
   * @return the value of the <code>background-color</code> property.
   * Returns <code>null</code> if the background color is transparent.
   */
  public Color getBackgroundColor() {
    Color bgColor = (Color) stylesArray[KEY_BACKGROUND_COLOR];
    if (bgColor != null) {
      return bgColor;
    }
    //Default value...
    return null;
  }

  /**
   * This property sets the background image of an element. When setting a background image, 
   * authors should also specify a background color that will be used when the image is unavailable. 
   * When the image is available, it is rendered on top of the background color. 
   * (Thus, the color is visible in the transparent parts of the image).
   * 
   * @return the value of the <code>background-image</code> property.
   * 
   * Returns <code>null</code> if the background image is not defined.  
   */
  public URIContent getBackgroundImage() {
    URIContent bgImage = (URIContent) stylesArray[KEY_BACKGROUND_IMAGE];
    if (bgImage != null) {
      return bgImage;
    }
    //Default value...
    return null;
  }

  /**
   * If a background image is specified, this property specifies whether the image is repeated (tiled), 
   * and how. All tiling covers the content, padding and border areas of a box.
   *  
   * @return One of CSS.
   */
  public String getBackgroundRepeat(){
    String bgRepeat = (String) stylesArray[KEY_BACKGROUND_REPEAT];
    if (bgRepeat != null) {
      return bgRepeat;
    }
    //Default value...
    return CSS.REPEAT;
  }


  /**
   * If a background position is specified, this property indicates the position of the background image 
   * relative to its box. 
   *  
   * @return A position, or <code>null</code>.
   */
  public BackgroundPosition getBackgroundPosition(){
    BackgroundPosition bgPosition = (BackgroundPosition) stylesArray[KEY_BACKGROUND_POSITION];
    if (bgPosition != null) {
      return bgPosition;
    }
    //Default value...
    return null;
  }
  

  
  /**
   * Obtain the color for full-tags background.
   * 
   * @return the value of the <code>-oxy-tags-background-color</code> property.
   * Returns <code>null</code> if no background color was defined in css.
   */
  public Color getTagsBackgroundColor() {
    Color bgColor = (Color) stylesArray[KEY_TAGS_BACKGROUND_COLOR];
    if (bgColor != null) {
      return bgColor;
    }
    
    return null;
  }
  
  /**
   * Obtain the color for full-tags background.
   * 
   * @return the value of the <code>-oxy-tags-background-color</code> property.
   * Returns <code>null</code> if no background color was defined in css.
   */
  public Color getTagsColor() {
    Color foregroundColor = (Color) stylesArray[KEY_TAGS_COLOR];
    if (foregroundColor != null) {
      return foregroundColor;
    }
    
    return null;
  }

  /**
   * @return the value of the <code>borderBottomColor</code> property.
   */
  public Color getBorderBottomColor() {
    Color borderColor = (Color) stylesArray[KEY_BORDER_BOTTOM_COLOR];
    if (borderColor != null) {
      return borderColor;
    }
    //Default value...
    return getColor();
  }

  /**
   * @return the value of the <code>borderBottomStyle</code> property.
   */
  public String getBorderBottomStyle() {
    String btStyle = (String) stylesArray[KEY_BORDER_BOTTOM_STYLE];
    if (btStyle != null) {
      return btStyle;
    }
    //Default value...
    return CSS.NONE;
  }

  /**
   * @return the value of the <code>borderLeftColor</code> property.
   */
  public Color getBorderLeftColor() {
    Color borderColor = (Color) stylesArray[KEY_BORDER_LEFT_COLOR];
    if (borderColor != null) {
      return borderColor;
    }
    //Default value...
    return getColor();
  }

  /**
   * @return the value of the <code>borderLeftStyle</code> property.
   */
  public String getBorderLeftStyle() {
    String btStyle = (String) stylesArray[KEY_BORDER_LEFT_STYLE];
    if (btStyle != null) {
      return btStyle;
    }
    //Default value...
    return CSS.NONE;
  }

  /**
   * @return the value of the <code>borderRightColor</code> property.
   */
  public Color getBorderRightColor() {
    Color borderColor = (Color) stylesArray[KEY_BORDER_RIGHT_COLOR];
    if (borderColor != null) {
      return borderColor;
    }
    //Default value...
    return getColor();
  }

  /**
   * @return the value of the <code>borderRightStyle</code> property.
   */
  public String getBorderRightStyle() {
    String btStyle = (String) stylesArray[KEY_BORDER_RIGHT_STYLE];
    if (btStyle != null) {
      return btStyle;
    }
    //Default value...
    return CSS.NONE;
  }

  /**
   * @return the value of the <code>borderTopColor</code> property.
   */
  public Color getBorderTopColor() {
    Color borderColor = (Color) stylesArray[KEY_BORDER_TOP_COLOR];
    if (borderColor != null) {
      return borderColor;
    }
    //Default value...
    return getColor();
  }

  /**
   * @return the value of the <code>borderTopStyle</code> property.
   */
  public String getBorderTopStyle() {
    String btStyle = (String) stylesArray[KEY_BORDER_TOP_STYLE];
    if (btStyle != null) {
      return btStyle;
    }
    //Default value...
    return CSS.NONE;
  }

  /**
   *@return the value of the <code>color</code> property.
   */
  public Color getColor() {
    Color color = (Color) stylesArray[KEY_FOREGROUND_COLOR];
    if (color != null) {
      return color;
    }
    // Default value...
    return Color.COLOR_BLACK;
  }

  /**
   * @return a <code>List</code> of <code>ContentPart</code> objects
   * representing the <code>content</code> property.
   */
  public StaticContent[] getMixedContent() {
    StaticContent[] mixedContent = (StaticContent[]) stylesArray[KEY_MIXED_CONTENT];
    if (mixedContent != null) {
      return mixedContent;
    }
    // Default value...
    return EMPTY_CONTENT;
  }
  
  /**
   * First it looks at the KEY_IMPOSED_DISPLAY property. If it is not set it looks at the 
   * KEY_DISPLAY property.
   * 
   * @return the value of the <code>display</code> property or CSS.INLINE if nothing is 
   * specified.
   */
  public String getDisplay() {
    String toReturn = CSS.INLINE;
    String btStyle = (String) stylesArray[KEY_IMPOSED_DISPLAY];
    if (btStyle == null) {
      // If a display is not imposed...
      btStyle = (String) stylesArray[KEY_DISPLAY];
      if (btStyle != null) {
        toReturn = btStyle;
      }
    } else {
      toReturn = btStyle;
    }
    //Default value...
    return toReturn;
  }
  
  /**
   * Check if the display has 'morph' or '-oxy-morph' value.
   * 
   * @return <code>true</code> for morph display.
   */
  public boolean isMorphDisplay() {
    String display = getDisplay();
    return CSS.MORPH.equals(display) || CSS.OXY_MORPH.equals(display);
  }

  /**
   * @return <code>true</code> if the associated element is invisible.
   */
  public boolean isInvisible() {
    String display = getDisplay();
    return CSS.NONE.equals(display) || CSS.DISPLAY_IGNORED.equals(display);
  }

  /**
   * @return <code>true</code> if the associated element is a list item.
   */
  public boolean isListItem() {
    return CSS.LIST_ITEM.equals(getDisplay());
  }

  /**
   * @return the value of the <code>font</code> property. Never null.
   */
  public Font getFont() {
    return (Font) stylesArray[KEY_FONT];
  }

  /**
   * @return the value of the <code>fontWeight</code> property.
   */
  public int getFontWeight() {
    Integer fontWeight = (Integer) stylesArray[KEY_FONT_WEIGHT];
    if (fontWeight != null) {
      return fontWeight.intValue();
    }

    //Default value...
    return FONT_WEIGHT_NORMAL;
  }

  /**
   * Sets the distance between lines: normal number length %
   * @param fm Font metrics for this style's font
   * @return the value of the <code>lineHeight</code> property.
   */
  public int getLineHeight(FontMetrics fm) {
    int lh = -1;
    Float multiplier = (Float) stylesArray[KEY_LINE_HEIGHT_MULTIPLIER];
    if (multiplier != null) {
      //If multiplier is specified, use it.
      lh = Math.round(fm.getHeight() * multiplier.floatValue());
    } else {
      lh = fm.getHeight();
    }
    return lh;
  }

  /**
   * @return the value of the <code>listStyleType</code> property.
   */
  public String getListStyleType() {
    String listStyleType = (String) stylesArray[KEY_LIST_STYLE_TYPE];
    if (listStyleType != null) {
      return listStyleType;
    }

    //Default value...
    return CSS.DISC;
  }

  /**
   * Aligns the text in an element: left right center justify.
   * 
   * @return the value of the <code>textAlign</code> property.
   */
  public String getTextAlign() {
    String textAlign = (String) stylesArray[KEY_TEXT_ALIGN];
    if (textAlign != null) {
      return textAlign;
    }

    //Default value...
    return CSS.LEFT;
  }

  /**
   * @return true if this element has the display property set to CSS.INLINE and thus it must be 
   * rendered as an inline block. Note that if this method returns false it doesn't necessarily 
   * mean that the element must be rendered as a block. The isInvisible() should be taken into 
   * account as well. See LayoutUtils.isInvisible() 
   */
  public boolean isInline() {
    return CSS.INLINE.equals(getDisplay());
  }

  /**
   * @return true if this element is table-formatted, or false otherwise.
   */
  public boolean isTableRowGroup() {
    return getDisplay().equals(CSS.TABLE_ROW_GROUP);
  }

  /**
   * @return true if this element is table-formatted, or false otherwise.
   */
  public boolean isTableRow() {
    return getDisplay().equals(CSS.TABLE_ROW);
  }

  /**
   * @return the value of border-bottom-width
   */
  public int getBorderBottomWidth() {
    Integer borderBottomWidth = (Integer) stylesArray[KEY_BORDER_BOTTOM_WIDTH];
    if (borderBottomWidth != null) {
      return borderBottomWidth.intValue();
    }

    // Default value.
    return BORDER_WIDTH_MEDIUM;
  }

  /**
   * @return the value of border-left-width
   */
  public int getBorderLeftWidth() {
    Integer borderLeftWidth = (Integer) stylesArray[KEY_BORDER_LEFT_WIDTH];
    if (borderLeftWidth != null) {
      return borderLeftWidth.intValue();
    }

    // Default value.
    return BORDER_WIDTH_MEDIUM;
  }

  /**
   * @return the value of border-right-width
   */
  public int getBorderRightWidth() {
    Integer borderRightWidth = (Integer) stylesArray[KEY_BORDER_RIGHT_WIDTH];
    if (borderRightWidth != null) {
      return borderRightWidth.intValue();
    }

    // Default value.
    return BORDER_WIDTH_MEDIUM;
  }

  /**
   * @return the value of border-top-width
   */
  public int getBorderTopWidth() {
    Integer borderTopWidth = (Integer) stylesArray[KEY_BORDER_TOP_WIDTH];
    if (borderTopWidth != null) {
      return borderTopWidth.intValue();
    }

    // Default value.
    return BORDER_WIDTH_MEDIUM;
  }

  /**
   * @return the value of margin-bottom
   */
  public RelativeLength getMarginBottom() {
    RelativeLength marginBottom = (RelativeLength) stylesArray[KEY_MARGIN_BOTTOM];
    if (marginBottom != null) {
      return marginBottom;
    }

    // Default value.
    return ZERO_RELATIVE_LENGTH;
  }

  /**
   * @return the value of margin-left
   */
  public RelativeLength getMarginLeft() {
    RelativeLength marginLeft = (RelativeLength) stylesArray[KEY_MARGIN_LEFT];
    if (marginLeft != null) {
      return marginLeft;
    }

    // Default value.
    return ZERO_RELATIVE_LENGTH;
  }

  /**
   * @return the value of margin-right
   */
  public RelativeLength getMarginRight() {
    RelativeLength marginRight = (RelativeLength) stylesArray[KEY_MARGIN_RIGHT];
    if (marginRight != null) {
      return marginRight;
    }

    // Default value.
    return ZERO_RELATIVE_LENGTH;
  }

  /**
   * @return the value of margin-top
   */
  public RelativeLength getMarginTop() {
    RelativeLength marginTop = (RelativeLength) stylesArray[KEY_MARGIN_TOP];
    if (marginTop != null) {
      return marginTop;
    }

    // Default value.
    return ZERO_RELATIVE_LENGTH;
  }

  /**
   * @return the value of padding-bottom
   */
  public RelativeLength getPaddingBottom() {
    RelativeLength paddingBottom = (RelativeLength) stylesArray[KEY_PADDING_BOTTOM];
    if (paddingBottom != null) {
      return paddingBottom;
    }

    // Default value.
    return ZERO_RELATIVE_LENGTH;
  }

  /**
   * @return the value of padding-left
   */
  public RelativeLength getPaddingLeft() {
    RelativeLength paddingLeft = (RelativeLength) stylesArray[KEY_PADDING_LEFT];
    if (paddingLeft != null) {
      return paddingLeft;
    }

    // Default value.
    return ZERO_RELATIVE_LENGTH;
  }

  /**
   * @return the value of padding-right
   */
  public RelativeLength getPaddingRight() {
    RelativeLength paddingRight = (RelativeLength) stylesArray[KEY_PADDING_RIGHT];
    if (paddingRight != null) {
      return paddingRight;
    }

    // Default value.
    return ZERO_RELATIVE_LENGTH;
  }

  /**
   * @return the value of padding-top
   */
  public RelativeLength getPaddingTop() {
    RelativeLength paddingTop = (RelativeLength) stylesArray[KEY_PADDING_TOP];
    if (paddingTop != null) {
      return paddingTop;
    }

    // Default value.
    return ZERO_RELATIVE_LENGTH;
  }

  /**
   * Checks if the display mode indicated to be a table.
   * 
   * @return True if it is.
   */
  public boolean isTable() {
    String display = getDisplay();
    return CSS.TABLE.equals(display) || CSS.INLINE_TABLE.equals(display);
  }

  /**
   * @return The whitespace value.
   */
  public String getWhitespace() {
    String whiteSpace = (String) stylesArray[KEY_WHITESPACE];
    if(whiteSpace != null) {
      return whiteSpace;
    }
    return CSS.NORMAL;
  }
  
  /**
   * @return The whitespace value.
   */
  public String getDirectWhitespace() {
    return (String) stylesArray[KEY_DIRECT_WHITESPACE];
  }

  /**
   * Checks if the display mode indicated to be a table group header.
   * 
   * @return True if it is.
   */
  public boolean isTableHeaderGroup() {
    return getDisplay().equals(CSS.TABLE_HEADER_GROUP);
  }

  /**
   * Checks if the display mode indicated to be a table group footer.
   * 
   * @return True if it is.
   */
  public boolean isTableFooterGroup() {
    return getDisplay().equals(CSS.TABLE_FOOTER_GROUP);
  }

  /**
   * Checks if the display mode indicated to be a table column.
   * 
   * @return True if it is.
   */

  public boolean isTableColumn() {
    return getDisplay().equals(CSS.TABLE_COLUMN);
  }

  /**
   * Checks if the display mode indicated to be a table group of columns.
   * 
   * @return True if it is.
   */
  public boolean isTableColumnGroup() {
    return getDisplay().equals(CSS.TABLE_COLUMN_GROUP);
  }

  /**
   * @return The specified width, as a relative length, or null if not specified by the CSS.
   */
  public RelativeLength getWidth() {
    RelativeLength width = (RelativeLength) stylesArray[KEY_WIDTH];
    if (width != null) {
      return width;
    }
    //Default value...
    return null;
  }
  
  /**
   * @return The specified height, as a relative length, or null if not specified by the CSS.
   */
  public RelativeLength getHeight() {
    RelativeLength height = (RelativeLength) stylesArray[KEY_HEIGHT];
    if (height != null) {
      return height;
    }
    //Default value...
    return null;
  }

  /**
   * @return The minimum width, or null if not specified.
   */
  public RelativeLength getMinWidth() {
    RelativeLength minWidth = (RelativeLength) stylesArray[KEY_MIN_WIDTH];
    if (minWidth != null) {
      return minWidth;
    }
    //Default value...
    return null;
  }

  /**
   * @return The maximum width, or null if not specified.
   */
  public RelativeLength getMaxWidth() {
    RelativeLength maxWidth = (RelativeLength) stylesArray[KEY_MAX_WIDTH];
    if (maxWidth != null) {
      return maxWidth;
    }
    //Default value...
    return null;
  }

  /**
   * @return true if the view is editable
   */
  public boolean isEditable() {
    Boolean editable = (Boolean) stylesArray[KEY_EDITABLE];
    if (editable != null) {
      return editable.booleanValue();
    }
    //Default value...
    return true;
  }

  /**
   * @return The horizontal border spacing between table cells. 
   */
  public int getHorizontalBorderSpacing() {
    return getBorderSpacing()[0];
  }

  /**
   * @return The vertical border spacing between table cells. 
   */
  public int getVerticalBorderSpacing() {
    return getBorderSpacing()[1];
  }

  /**
   * @return The value of 'border-spacing' property. 
   * An array with two integers for horizontal and vertical spacing.
   */
  private int[] getBorderSpacing() {
    int[] borderSpacing = (int[]) stylesArray[KEY_BORDER_SPACING];
    if (borderSpacing != null) {
      return borderSpacing;
    }
    return EMPTY_BORDER_SPACING;
  }

  /**
   * Set a property to the map.
   * 
   * @param property The key of the property to be set. It can be one of the keys defined in
   * this class.
   * @param value The value of the property. 
   * Accepted values depend on the property and can have as type {@link String}, 
   * {@link Color}, {@link CSSCounter} or {@link CSSCounterIncrement}.
   * The actual values are the Java representations of the ones accepted in the CSS specification.
   * Read more about the CSS support in the Developer Guide.
   */
  public void setProperty(int property, Object value) {
    stylesArray[property]= value;
  }

  /**
   * @param property The property name.
   * @return The value of the property.
   */
  public Object getProperty(int property) {
    return stylesArray[property];
  }
  
  /**
   * Copy all additional styles (which are not set in the current styles)
   * from a source.
   * 
   * @param styles Source of the copy operation
   * @param onlyInheritable <code>true</code> to copy only inheritable styles. If <code>false</code>
   * all styles will be copied regardless if they are inheritable or not.
   */
  void copyAdditionalStylesFrom(Styles styles, boolean onlyInheritable) {
    if (styles != null) {
      Object[] source = styles.stylesArray;
      for (int i = 0; i < source.length; i++) {
        boolean copyToMap = stylesArray[i] == null 
            && i != KEY_COUNTER_RESET
            && i != KEY_COUNTER_INCREMENT
        //EXM-14130 The mixed content should not be inherited.
            && i != KEY_MIXED_CONTENT
        //EXM-18564 The filtered state should not get copied.
            && i != KEY_FILTERED_OUT
            && (!onlyInheritable || isInheritable(i));
        if (!copyToMap) {
          //EXM-9238 Display None should always overwrite the display in the map.
          //It is stronger and it needs to propagate to all children from the parent.
          if (KEY_DISPLAY == i) {
            String value = (String) source[i];
            if (CSS.NONE.equals(value)) {
              copyToMap = true;
            }
          }
        }
        if (copyToMap) {
          stylesArray[i] = source[i];
        }
      }
      
      //EXM-26093 Force empty border, the same is done from the Stylesheet part.
      if((CSS.NONE.equals(getBorderBottomStyle())
          || CSS.HIDDEN.equals(getBorderBottomStyle()))) {
        stylesArray[KEY_BORDER_BOTTOM_WIDTH] = NumberParserUtil.valueOfInteger(0);
      } 
      if((CSS.NONE.equals(getBorderTopStyle())
          || CSS.HIDDEN.equals(getBorderTopStyle()))) {
        stylesArray[KEY_BORDER_TOP_WIDTH] = NumberParserUtil.valueOfInteger(0);
      } 
      if((CSS.NONE.equals(getBorderLeftStyle())
          || CSS.HIDDEN.equals(getBorderLeftStyle()))) {
        stylesArray[KEY_BORDER_LEFT_WIDTH] = NumberParserUtil.valueOfInteger(0);
      }  
      if((CSS.NONE.equals(getBorderRightStyle())
          || CSS.HIDDEN.equals(getBorderRightStyle()))) {
        stylesArray[KEY_BORDER_RIGHT_WIDTH] = NumberParserUtil.valueOfInteger(0);
      }
    }
  }
  
  /**
   * The styles that can't be copied from a parent to a child.
   */
  private static final ArrayList<Integer> UN_INHERITABLE_STYLES = new ArrayList<Integer>();

  static {
    UN_INHERITABLE_STYLES.add(KEY_DISPLAY);
    UN_INHERITABLE_STYLES.add(KEY_COUNTER_RESET);
    UN_INHERITABLE_STYLES.add(KEY_COUNTER_INCREMENT);
    // EXM-15034 Fold properties must not be inherited.
    UN_INHERITABLE_STYLES.add(KEY_FOLDABLE);
    UN_INHERITABLE_STYLES.add(KEY_FOLDED);
    //EXM-14130 The mixed content should not be inherited.
    UN_INHERITABLE_STYLES.add(KEY_MIXED_CONTENT);
    //EXM-18564 The filtered state should not get copied.
    UN_INHERITABLE_STYLES.add(KEY_FILTERED_OUT);
    UN_INHERITABLE_STYLES.add(KEY_SHOW_PLACEHOLDER); 
    UN_INHERITABLE_STYLES.add(KEY_PLACEHOLDER_CONTENT); 
    UN_INHERITABLE_STYLES.add(KEY_MARGIN_BOTTOM); 
    UN_INHERITABLE_STYLES.add(KEY_MARGIN_LEFT); 
    UN_INHERITABLE_STYLES.add(KEY_MARGIN_RIGHT); 
    UN_INHERITABLE_STYLES.add(KEY_MARGIN_TOP); 
    UN_INHERITABLE_STYLES.add(KEY_PADDING_BOTTOM); 
    UN_INHERITABLE_STYLES.add(KEY_PADDING_LEFT); 
    UN_INHERITABLE_STYLES.add(KEY_PADDING_RIGHT); 
    UN_INHERITABLE_STYLES.add(KEY_PADDING_TOP); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_BOTTOM_COLOR); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_BOTTOM_STYLE); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_BOTTOM_WIDTH); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_LEFT_COLOR); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_LEFT_STYLE); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_LEFT_WIDTH); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_RIGHT_COLOR); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_RIGHT_STYLE); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_RIGHT_WIDTH); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_TOP_COLOR); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_TOP_STYLE); 
    UN_INHERITABLE_STYLES.add(KEY_BORDER_TOP_WIDTH); 
    
    UN_INHERITABLE_STYLES.add(KEY_BACKGROUND_COLOR); 
    UN_INHERITABLE_STYLES.add(KEY_BACKGROUND_IMAGE); 
    UN_INHERITABLE_STYLES.add(KEY_BACKGROUND_REPEAT); 
    UN_INHERITABLE_STYLES.add(KEY_BACKGROUND_POSITION); 
    
  }

  /**
   * Check if the given property should be copied from parent to child. A property
   * can be copied if is marked as being inheritable.
   * 
   * @return <code>true</code> if the given property can be copied.
   */
  static boolean isInheritable(int property) {
    return !UN_INHERITABLE_STYLES.contains(property);
  }

  /**
   * @return The value of the 'visibility' 
   */
  public String getVisibility() {
    String visibility = CSS.VISIBILITY_VISIBLE;
    String propVal = (String) stylesArray[KEY_VISIBITY];
    if (propVal != null) {
      visibility = propVal;
    }
    return visibility;
  }
  
  /**
   * @return The value of the 'show-placeholder' property. 
   */
  public String getShowPlaceholders() {
    return (String) stylesArray[KEY_SHOW_PLACEHOLDER];
  }
  
  /**
   * @return The value of the 'placeholder-content' property. 
   */
  public String getPlaceholderContent() {
    return (String) stylesArray[KEY_PLACEHOLDER_CONTENT];
  }

  /**
   * @return Return true if table empty cells are shown. 
   * See http://www.w3.org/TR/CSS21/tables.html#empty-cells.
   */
  public boolean getShowEmptyCells() {
    boolean showEmptyCells = true;
    Boolean propVal = (Boolean) stylesArray[KEY_EMPTY_CELLS];
    if (propVal != null) {
      showEmptyCells = propVal.booleanValue();
    }
    return showEmptyCells;
  }

  /**
   * @return true if this element is foldable
   */
  public boolean isFoldable() {
    Boolean foldable = (Boolean) stylesArray[KEY_FOLDABLE];
    if (foldable != null) {
      return foldable.booleanValue();
    }

    //Default value
    return false;
  }
  
  /**
   * @return <code>true</code> if this element is folded
   */
  public boolean isFolded() {
    Boolean folded = (Boolean) stylesArray[KEY_FOLDED];
    // false is the default value.
    return folded != null ? folded : false;
  }

  /**
   * @return The name of the child which is non foldable
   */
  public String[] getNonFoldableChildName() {
    String[] childrenName = (String[]) stylesArray[KEY_NON_FOLDABLE_CHILD_NAME];
    if (childrenName != null) {
      return childrenName;
    }
    //Default value
    return EMPTY_STRING_ARRAY;
  }

  /**
  * Checks if the display mode indicated to be a table caption.
  * 
  * @return True if it is.
  */
  public boolean isTableCaption() {
    return getDisplay().equals(CSS.TABLE_CAPTION);
  }

  /**
   * @return <code>true</code> if element is specified as INLINE in CSS source.
   */
  public boolean isInlineInCSS() {
    String display = (String) stylesArray[KEY_DISPLAY];
    return display == null || CSS.INLINE.equals(display);
  }

  /**
   * @return The vertical align.
   */
  public String getVerticalAlign() {
    String va = (String) stylesArray[KEY_VERTICAL_ALIGN];
    return va != null ? va : CSS.BASELINE;
  }

  /**
   * @return The value of 'text-decoration' property.
   */
  public String[] getTextDecoration() {
    String[] va = (String[]) stylesArray[KEY_TEXT_DECORATION];
    return va != null ? va : new String[] { CSS.NONE };
  }

  /**
   * @return The color to be used for paint text decoration.
   */
  public Color getTextDecorationColor() {
    Color color = (Color) stylesArray[KEY_TEXT_DECORATION_COLOR];
    if (color != null) {
      return color;
    }
    // Default value...
    return Color.COLOR_BLACK;
  }

  /**
   * @return The counters declared on this element or empty array if none.
   */
  public CSSCounter[] getCounters() {
    CSSCounter[] counters = (CSSCounter[]) stylesArray[KEY_COUNTER_RESET];
    return counters == null ? EMPTY_COUNTERS : counters;
  }

  /**
   * @return The 'counters increment' declared on this element or empty array if none.
   */
  public CSSCounterIncrement[] getCountersIncrement() {
    CSSCounterIncrement[] counters = (CSSCounterIncrement[]) stylesArray[KEY_COUNTER_INCREMENT];
    return counters == null ? EMPTY_COUNTERS_INCREMENT : counters;
  }

  /**
   * Check if is a style from a table but not a cell.
   * 
   * @return <code>true</code> if the current style is a one of
   *  CSS.TABLE, CSS.INLINE_TABLE, CSS.TABLE_ROW, CSS.TABLE_ROW_GROUP, 
   *  CSS.TABLE_FOOTER_GROUP, CSS.TABLE_HEADER_GROUP
   */
  public boolean isInTable() {
    return (isTable() ||
        isTableRow() ||
        isTableRowGroup() || 
        isTableFooterGroup() ||
        isTableHeaderGroup());
  }

  /**
   * @return The value of the 'link' property. Null if property was absent. 
   */
  public String getLinkURL() {
    String linkURL = (String) stylesArray[KEY_LINK_URL];
    return linkURL;
  }

  /**
   * @return Returns the value of the property 'display-tags'. 
   * The possible values are:
   * <ul>
   *   <li>CSS.DEFAULT - display tag markers depending on the current display mode;</li>
   *   <li>CSS.NONE - tag markers will not be shown.</li>
   * </ul>
   * The default value is CSS.DEFAULT.
   */
  public String getDisplayTags() {
    String displayTags = (String) stylesArray[KEY_DISPLAY_TAGS];
    return displayTags != null ? displayTags : CSS.DEFAULT;
  }
  
  /**
   * @return Returns the value of the property 'text-transform'. 
   * The possible values are:
   * <ul>
   *   <li>CSS.CAPITALIZE
   *   <li>CSS.UPPERCASE
   *   <li>CSS.LOWERCASE
   * </ul>
   * The default value is null
   */
  public String getTextTransform() {
    return (String) stylesArray[KEY_TEXT_TRANSFORM];
  }
  
  /**
   * @return true if a node should be represented with faded colors.
   */
  public boolean isFilteredOut() {
    Boolean filtered = (Boolean) stylesArray[KEY_FILTERED_OUT];
    return filtered != null ? filtered.booleanValue() : false;
  }
  
  /**
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (obj instanceof Styles) {
      Styles s = (Styles) obj;
      for (int i = 0; i < stylesArray.length; i++) {
        if (stylesArray[i] instanceof Object[] && s.stylesArray[i] instanceof Object[]) {
          if (!Equaler.verifyArrayEquals((Object[]) stylesArray[i], (Object[]) s.stylesArray[i])) {
            return false;
          }
        } else if (!Equaler.verifyEquals(stylesArray[i], s.stylesArray[i])) {
          return false;
        }
      }
      return true;
    }
    return false;
  }
  
  /**
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    return hash(23, stylesArray);
  }

  /**
   * Hash for objects.
   * <code>aObject</code> is a possibly-null object field, and possibly an array.
   *
   * If <code>aObject</code> is an array, then each element may be a primitive
   * or a possibly-null object.
   */
  private static int hash(int aSeed, Object aObject) {
    int result = aSeed;
    if (aObject == null) {
      result = hash(result, 0);
    } else if (!isArray(aObject)) {
      result = hash(result, aObject.hashCode());
    } else {
      int length = Array.getLength(aObject);
      for (int idx = 0; idx < length; ++idx) {
        Object item = Array.get(aObject, idx);
        if (item != null) {
          // recursive call
          result = hash(result, item);
        }
      }
    }
    return result;
  }

  /**
   * Hash for ints.
   */
  private static int hash(int aSeed, int aInt) {
    return 37 * aSeed + aInt;
  }

  private static boolean isArray(Object aObject) {
    return aObject.getClass().isArray();
  }
  
  /**
   * Shallow clone.
   */
  @Override
  public Object clone() {
    Styles clonedStyles = new Styles();
    clonedStyles.stylesArray = new Object[stylesArray.length];
    System.arraycopy(stylesArray, 0, clonedStyles.stylesArray, 0, stylesArray.length);
    return clonedStyles;
  }
  
  /**
   * Verify if styles can be cached. Can be cached if properties values does not depends by attributes or document structure.
   *  
   * @return True if styles can be cached.
   */
  public boolean canBeCached() {
    return canBeCached 
      && getCounters().length == 0 
      // If a content property exists it might contain counters. Rather than 
      // checking if counters exist (which might be time consuming) we prefer 
      // not to cache no matter is counters exist or not.
      && getMixedContent().length == 0;
  }

  /**
   * Sets if those styles can be shared between two or more nodes.
   * 
   * @param canBeCached True when be shared between two or more nodes.
   */
  public void setCanBeCached(boolean canBeCached) {
    this.canBeCached = canBeCached;
  }
  
  /**
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString() {
    String val = "STYLES:\n";
    
    for (int i = 0; i < stylesArray.length; i++) {
      if(stylesArray[i] != null) {
        val += "key " + i + " val:" + stylesArray[i] + "\n";
      }
    }
    return val;
  }
}