/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.api.component;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.document.DocumentPositionedInfo;
import ro.sync.ecss.extensions.api.component.listeners.OpenURLHandler;
import ro.sync.exml.editor.AbstractEditor;
import ro.sync.exml.editor.AutomaticallyValidationManager;
import ro.sync.exml.editor.EditorHighlightPaintInfo;
import ro.sync.exml.editor.EditorManagerAdapter;
import ro.sync.exml.editor.PersistentEditor;
import ro.sync.exml.editor.validate.ValidationEditorInterface;
import ro.sync.util.Equaler;

/**
 * The base component editor manager
 */
@API(type=APIType.INTERNAL, src=SourceType.PRIVATE)
public class BaseComponentEditorManager extends EditorManagerAdapter { 
  /**
   * The validation manager
   */
  private AutomaticallyValidationManager validationManager;
  
  /**
   * A handler notified when an URL should be opened.
   */
  private OpenURLHandler openURLHandler = null;

  /**
   * All provider
   */
  private final List allProviders;

  /**
   * The last focused component
   */
  private InternalComponentProvider lastFocusedProvider;
  
  /**
   * Constructor.
   *
   * @param allProviders All component providers.
   */
  public BaseComponentEditorManager(List allProviders) {
    this.allProviders = allProviders;
    validationManager = new AutomaticallyValidationManager(this);
  }
  
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#getEditorForURL(java.net.URL)
   */
  @Override
  public PersistentEditor getEditorForURL(URL url) {
    AbstractEditor[] allEditors = getAllEditors();
    for (int i = 0; i < allEditors.length; i++) {
      if(Equaler.verifyEquals(url, allEditors[i].getEditorLocation())) {
        return allEditors[i];
      }
    }
    return null;
  }
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#findValidationEditor(java.net.URL)
   */
  @Override
  public ValidationEditorInterface findValidationEditor(URL url) {
    return (ValidationEditorInterface) getEditorForURL(url);
  }
  /**
   * @return Returns the validationManager.
   */
  @Override
  public AutomaticallyValidationManager getAutomaticallyValidationManager() {
    return validationManager;
  }
  /**
   * EXM-18738
   * @see ro.sync.exml.editor.EditorManagerAdapter#getSelectedEditor()
   */
  @Override
  public AbstractEditor getSelectedEditor() {
    if(lastFocusedProvider != null) {
      return (AbstractEditor) lastFocusedProvider.getEditorKey();
    } else {
      AbstractEditor[] allEditors = getAllEditors();
      return allEditors != null && allEditors.length > 0 ? allEditors[allEditors.length - 1] : null;
    }
  }
  /**
   * EXM-18738
   * @see ro.sync.exml.editor.EditorManagerAdapter#getAllEditors()
   */
  @Override
  public AbstractEditor[] getAllEditors() {
    List<AbstractEditor> allEditors = new ArrayList<AbstractEditor>();
    for (int i = 0; i < allProviders.size(); i++) {
      InternalComponentProvider provider = (InternalComponentProvider) allProviders.get(i);
      allEditors.add((AbstractEditor) provider.getEditorKey());
    }
    return allEditors.toArray(new AbstractEditor[0]);
  }
  
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#getAllOpenedURLs()
   */
  @Override
  public URL[] getAllOpenedURLs() {
    List<URL> allEditorURLs = new ArrayList<URL>();
    for (int i = 0; i < allProviders.size(); i++) {
      InternalComponentProvider provider = (InternalComponentProvider) allProviders.get(i);
      allEditorURLs.add(((AbstractEditor) provider.getEditorKey()).getEditorLocation());
    }
    return allEditorURLs.toArray(new URL[0]);
  }

  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#open(java.net.URL, boolean, java.lang.String)
   */
  @Override
  public Object open(URL toLoad0, boolean isNew, String contentType) throws IOException {
    //Delegate to the handler
    if(openURLHandler != null) {
      openURLHandler.handleOpenURL(toLoad0);
    }
    return null;
  }
  
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#open(java.net.URL, java.lang.String, boolean, boolean)
   */
  @Override
  public Object open(URL toLoad0, String contentType, boolean requestFocus, boolean ignore)
      throws IOException {
    //Delegate to the handler
    if(openURLHandler != null) {
      openURLHandler.handleOpenURL(toLoad0);
    }
    return null;
  }
  
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#open(java.net.URL)
   */
  @Override
  public Object open(URL url) throws IOException {
    //Delegate to the handler
    if(openURLHandler != null) {
      openURLHandler.handleOpenURL(url);
    }
    return null;
  }
  
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#open(java.net.URL, java.lang.String, java.lang.String)
   */
  @Override
  public Object open(URL url, String imposedInitialPage, String imposedContentType)
      throws IOException {
    //Delegate to the handler
    if(openURLHandler != null) {
      openURLHandler.handleOpenURL(url);
    }
    return null;
  }
  
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#openAndHighlightLocation(java.net.URL, ro.sync.document.DocumentPositionedInfo, java.lang.String, java.lang.String, ro.sync.exml.editor.EditorHighlightPaintInfo, boolean)
   */
  @Override
  public Object openAndHighlightLocation(URL fileURL, DocumentPositionedInfo dpInfo,
      String contentType, String highlightClass, EditorHighlightPaintInfo highlightPainterInfo,
      boolean moveCaretToStart) throws IOException {
    //Delegate to the handler
    if(openURLHandler != null) {
      openURLHandler.handleOpenURL(fileURL);
    }
    return null;
  }
  
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#openAndShowLocation(java.net.URL, ro.sync.document.DocumentPositionedInfo, boolean, java.lang.String)
   */
  @Override
  public Object openAndShowLocation(URL fileURL, DocumentPositionedInfo dpInfo,
      boolean requestFocus, String contentType) throws IOException {
    //Delegate to the handler
    if(openURLHandler != null) {
      openURLHandler.handleOpenURL(fileURL);
    }
    return null;
  }
  
  /**
   * @see ro.sync.exml.editor.EditorManagerAdapter#openOperationResult(java.net.URL)
   */
  @Override
  public Object openOperationResult(URL url) throws IOException {
    //Delegate to the handler
    if(openURLHandler != null) {
      openURLHandler.handleOpenURL(url);
    }
    return null;
  }
  
  /**
   * Set a handler which will be notified when an URL should be opened.
   * For example the user clicked a link in the Author page or double clicked in the DITA Maps manager.
   * 
   * @param openURLHandler The open URLs handler.
   */
  public void setOpenURLHandler(OpenURLHandler openURLHandler) {
    this.openURLHandler = openURLHandler;
  }

  /**
   * Set the last focused component
   * 
   * @param componentProvider The last focused component.
   */
  public void setLastFocusedComponent(InternalComponentProvider componentProvider) {
    if(this.lastFocusedProvider != null && this.lastFocusedProvider.getEditorKey() != null) {
      ((AbstractEditor) this.lastFocusedProvider.getEditorKey()).editorSelected(true, false);  
    }
    this.lastFocusedProvider = componentProvider;
    if(componentProvider != null && componentProvider.getEditorKey() != null) {
      ((AbstractEditor) componentProvider.getEditorKey()).editorSelected(true, false);
    }
  }

  /**
   * Remove references to diposed components.
   * 
   * @param provider The provider was disposed.
   */
  public void componentDisposed(InternalComponentProvider provider) {
    if(lastFocusedProvider == provider) {
      lastFocusedProvider = null;
    }
  }
}