/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.exml.view.graphics;

import java.util.StringTokenizer;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.annotations.obfuscate.SkipLevel;
import ro.sync.annotations.obfuscate.SkipObfuscate;
import ro.sync.util.Equaler;
import ro.sync.util.NumberParserUtil;
import ro.sync.util.NumberFormatException;

/**
 * Font common class.
 */
@API(type=APIType.EXTENDABLE, src=SourceType.PRIVATE)
@SkipObfuscate(classes = SkipLevel.PRIVATE, fields = SkipLevel.PRIVATE, methods = SkipLevel.PRIVATE)
public class Font {

  /**
   * The plain style constant.
   */
  public static final int PLAIN = 0;

  /**
   * The bold style constant.  This can be combined with the other style
   * constants (except PLAIN) for mixed styles.
   */
  public static final int BOLD = 1;

  /**
   * The italicized style constant.  This can be combined with the other
   * style constants (except PLAIN) for mixed styles.
   */
  public static final int ITALIC = 2;

  /**
   * Font name
   */
  private final String name;

  /**
   * Font style
   */
  private final int style;

  /**
   * Font size
   */
  private final int size;
  
  /**
   * The font hash
   */
  private String hash;
  
  /**
   * The font hash code
   */
  private int hashCode = -1;

  /**
   * Creates a new <code>Font</code> from the specified name, style and
   * point size.
   * @param name the font name.  This can be a logical font name or a
   * font face name. A logical name must be either: Dialog, DialogInput,
   * Monospaced, Serif, or SansSerif.  If <code>name</code> is
   * <code>null</code>, the <code>name</code> of the new
   * <code>Font</code> is set to the name "Default".
   * @param style the style constant for the <code>Font</code>
   * The style argument is an integer bitmask that may
   * be PLAIN, or a bitwise union of BOLD and/or ITALIC
   * (for example, ITALIC or BOLD|ITALIC).
   * @param size the point size of the <code>Font</code>
   */
  public Font(String name, int style, int size) {
    this.name = name;
    this.style = style;
    this.size = size;
  }

  /**
   * @return The name
   */
  public String getName() {
    return name;
  }

  /**
   * @return The size
   */
  public int getSize() {
    return size;
  }

  /**
   * @return The style
   */
  public int getStyle() {
    return style;
  }

  /**
   * Converts this <code>Font</code> object to a <code>String</code>
   * representation.
   * @return     a <code>String</code> representation of this 
   *    <code>Font</code> object.
   * @since      JDK1.0
   */
  @Override
  public String toString() {
    String  strStyle;

    if (isBold()) {
      strStyle = isItalic() ? "bolditalic" : "bold";
    } else {
      strStyle = isItalic() ? "italic" : "plain";
    }

    return getClass().getName() + "[name=" + name + ",style=" +
    strStyle + ",size=" + size + "]";
  }
  
  /**
   * Indicates whether or not this <code>Font</code> object's style is
   * PLAIN.
   * @return    <code>true</code> if this <code>Font</code> has a
   *      PLAIN sytle;
   *            <code>false</code> otherwise.
   * @see       java.awt.Font#getStyle
   * @since     JDK1.0
   */
  public boolean isPlain() {
    return style == 0;
  }

  /**
   * Indicates whether or not this <code>Font</code> object's style is
   * BOLD.
   * @return    <code>true</code> if this <code>Font</code> object's
   *      style is BOLD;
   *            <code>false</code> otherwise.
   * @see       java.awt.Font#getStyle
   * @since     JDK1.0
   */
  public boolean isBold() {
    return (style & BOLD) != 0;
  }

  /**
   * Indicates whether or not this <code>Font</code> object's style is
   * ITALIC.
   * @return    <code>true</code> if this <code>Font</code> object's
   *      style is ITALIC;
   *            <code>false</code> otherwise.
   * @see       java.awt.Font#getStyle
   * @since     JDK1.0
   */
  public boolean isItalic() {
    return (style & ITALIC) != 0;
  }
  
  /**
   * @return A hash of the font
   */
  public String hash() {
    if(hash ==  null) {
      hash = hash(name, style, size); 
    }
    return hash;
  }
  
  /**
   * Hash the font
   * 
   * @param name Name
   * @param style Style
   * @param size Size
   * @return The hash
   */
  public static final String hash(String name, int style, int size) {
    return name + "." + style + "." + size;
  }
  

  /**
   * Checks if the two colors have the same RGB value. The alpha is also taken into account.
   */
  @Override
  public boolean equals(Object obj) {
    boolean equals = false;
    if (obj instanceof Font) {
      Font otherFont = (Font) obj;
      equals = this.hashCode() == otherFont.hashCode() &&
        Equaler.verifyEquals(this.hash(), otherFont.hash());
    }
    return equals;
  }

  /**
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    if(hashCode == -1) {
      hashCode = hash().hashCode();
    }
    return hashCode;
  }
  
  /**
   *  Encodes a font as: "name,style,size"
   *
   *@param  font  The font to be encoded.
   *@return       The encoded value.
   */
  public static String encodeFont(Font font) {
    return font.getName() + "," + font.getStyle() + "," + font.getSize();
  }
  
  /**
   *  Decodes a font property value, having the format "name,style,size"
   *
   *@param  value  The value to be decoded.
   *@return        The font.
   */
  public static Font decodeFont(String value) {
    Font font = null;
    if (value != null && value.length() > 0) {
      StringTokenizer stk = new StringTokenizer(value, ",", false);
      if (stk.countTokens() == 3) {
        try {
          String name = stk.nextToken();
          int style = NumberParserUtil.parseInt(stk.nextToken());
          int size = NumberParserUtil.parseInt(stk.nextToken());
          font = new Font(name, style, size);
        } catch (NumberFormatException nfe) {
          //Keeps the default
        }
      }
    }
    return font;
  }
}
