/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.exml.view.graphics;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.annotations.obfuscate.SkipLevel;
import ro.sync.annotations.obfuscate.SkipObfuscate;

/**
 *  The graphics interface used to draw Author and Schema Diagram.
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PRIVATE)
@SkipObfuscate(classes = SkipLevel.PRIVATE, fields = SkipLevel.PRIVATE, methods = SkipLevel.PRIVATE)
public interface Graphics {

  /**
   * Icon code for text nodes.
   */
  int ICON_TEXT_TREE = 0;
  
  /**
   * Icon code for comment nodes.
   */
  int ICON_COMMENT_TREE = 1;
  
  /**
   * Icon code for processing instruction nodes.
   */
  int ICON_PROCESSING_INSTRUCTION_TREE = 2;
  
  /**
   * Image representing the top arrow of the expansion bar.
   */
  int IMAGE_EXPANSION_BAR_TOP = 5;
  
  /**
   * Image representing the bottom arrow of the expansion bar.
   */
  int IMAGE_EXPANSION_BAR_BOTTOM = 6;
  
  /**
   * Image representing the middle of the expansion bar gradient.
   */
  int IMAGE_EXPANSION_BAR_MIDDLE_GRADIENT = 7;

  /**
   * Image representing an arrow for the collapsed expansion bar.
   */
  int IMAGE_EXPANSION_BAR_COLLAPSED = 9;
  
  /**
   * Normal stroke
   */
  int STROKE_NORMAL = 0;
  
  /**
   * Dotted stroke
   */
  int STROKE_DOTTED = 1;

  /**
   * Dotted stroke
   */
  int STROKE_DASHED = 2;
  
  /**
   * Wavy stroke
   */
  int STROKE_WAVY = 3;
  
  /**
   * Dot-dash stroke
   */
  int STROKE_DOT_DASH = 4;
  
  /**
   * Dot-dot-dash stroke
   */
  int STROKE_DOT_DOT_DASH = 5;

  /**
   * A pattern of squares (of size 1).
   */
  int FILL_PATTERN_SQUARE = 1;

  /**
   * A pattern of big squares (of size 2).
   */
  int FILL_PATTERN_SQUARE_BIG = 2;
  
  /**
   * Set the new font.
   * 
   * @param font  The font to be used.  
   */
  void setFont(Font font);

  /**
   * Fills a closed polygon defined by arrays of <i>x</i> and <i>y</i> coordinates. 
   * 
   * @param  xPoints   a an array of <code>x</code> coordinates.
   * @param  yPoints   a an array of <code>y</code> coordinates.
   * @param  nPoints   a the total number of points.   
   */
  void fillPolygon(int xPoints[], int yPoints[], int nPoints);

  /**
   * Draw a closed polygon defined by arrays of <i>x</i> and <i>y</i> coordinates. 
   * 
   * @param  xPoints   a an array of <code>x</code> coordinates.
   * @param  yPoints   a an array of <code>y</code> coordinates.
   * @param  nPoints   a the total number of points.   
   */
  void drawPolygon(int[] xPoints, int[] yPoints, int nPoints);

  /**
   * Get current font. 
   * 
   * @return  The current font.
   */
  Font getFont();

  /**
   * Draw a string.
   *
   * @param text  The text to be drawn.
   * @param x The x coordinate.
   * @param y The y coordinate(top-left).
   */
  void drawString(String text, int x, int y);

  /**
   * Get the current color code.
   * 
   * @return  The current color code.
   */
  Color getDrawColor();

  /**
   * Draw a line from (x1, y1) to (x2, y2).
   * 
   * @param x1  x of first point.
   * @param y1  y of first point. 
   * @param x2  x of second point.
   * @param y2  y of second point.
   */
  void drawLine(int x1, int y1, int x2, int y2);

  /**
   * Fill a rect with (x,y) top-left corner.
   * 
   * @param x The x coordinate.
   * @param y The y coordinate.
   * @param width The rect width.
   * @param height  The rect height.
   */
  void fillRect(int x, int y, int width, int height);
  
  /**
   * Fill a rectangle.
   * 
   * @param rectangle The rectangle.
   */
  void fillRect(Rectangle rectangle);

  
  /**
   * Draw a rect with (x,y) top-left corner.
   * 
   * @param x The x coordinate.
   * @param y The y coordinate.
   * @param width The rect width.
   * @param height  The rect height.
   */
  void drawRect(int x, int y, int width, int height);
  
  /**
   * Fills the specified rounded corner rectangle.
   * 
   * @param       x the <i>x</i> coordinate of the rectangle to be filled.
   * @param       y the <i>y</i> coordinate of the rectangle to be filled.
   * @param       width the width of the rectangle to be filled.
   * @param       height the height of the rectangle to be filled.
   * @param       arcWidth the horizontal diameter 
   *                     of the arc at the four corners.
   * @param       arcHeight the vertical diameter 
   *                     of the arc at the four corners.
   */
  void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight);

  /**
   * Draws a polyline which is not closed if the last point differs from the first.
   * 
   * @param xpoints The xpoints array.
   * @param ypoints The ypoints array.
   * @param npoints The number of points.
   */
  void drawPolyline(int[] xpoints, int[] ypoints, int npoints);

  /**
   * draws an arc.
   * 
   * @param        x the <i>x</i> coordinate of the 
   *                    upper-left corner of the arc to be drawn.
   * @param        y the <i>y</i>  coordinate of the 
   *                    upper-left corner of the arc to be drawn.
   * @param        width the width of the arc to be drawn.
   * @param        height the height of the arc to be drawn.
   * @param        startAngle the beginning angle.
   * @param        arcAngle the angular extent of the arc, 
   *                    relative to the start angle.
   */
  void drawArc(int x, int y, int width, int height, int startAngle, int arcAngle);
  
  /**
   * Fills an arc.
   * 
   * @param        x the <i>x</i> coordinate of the 
   *                    upper-left corner of the arc to be drawn.
   * @param        y the <i>y</i>  coordinate of the 
   *                    upper-left corner of the arc to be drawn.
   * @param        width the width of the arc to be drawn.
   * @param        height the height of the arc to be drawn.
   * @param        startAngle the beginning angle.
   * @param        arcAngle the angular extent of the arc, 
   *                    relative to the start angle.
   */
  void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle);

  /**
   * Sets the line width for the stroke.
   * The line is solid, no dashes.
   * 
   * @param width The width value.
   * @param type The stroke type.
   */
  void setStroke(int width, int type);
      
  /**
   * @return the value of the stroke width.
   */
  int getStrokeWidth();

  /**
   * @return the value of the stroke type.
   */
  int getStrokeType();

  /**
   * @return the clipping rectangle.
   */
  Rectangle getClipBounds();

  /**
   * Set the color for fill methods.
   * 
   * @param colorCode The color code used by fill methods. 
   */
  void setFillColor(Color colorCode);
  
  /**
   * Set the color for draw methods.
   * 
   * @param colorCode The color code used by draw methods. 
   */
  void setDrawColor(Color colorCode);

  /**
   * 
   * Draw the icon at (x,y) coordinates.
   *
   * @param hostComponent Icon implementations may use the HostComponent argument to 
   *  get properties useful for painting, e.g. the foreground or background color.
   * @param iconCode  The icon code.
   * @param x The x coordinate.
   * @param y The y coordinate.
   */
  void drawIcon(Object hostComponent, int iconCode, int x, int y);
  
  /**
   * Draw an oval. 
   * 
   * @param x The x coordinate.
   * @param y The y coordinate.
   * @param width The width.
   * @param height The height. 
   */
  void drawOval(int x, int y, int width, int height);

  /**
   * Get the color code for fill shapes.
   * 
   * @return  The color code for fill shapes.
   */
  Color getFillColor();

  /**
   * Fill an oval with curent fill color. 
   * 
   * @param x The x coordinate.
   * @param y The y coordinate.
   * @param width The width.
   * @param height The height. 
   */
  void fillOval(int x, int y, int width, int height);

  /**
   * @param rect The rect to fill with a gradient
   * @param startColorCode Start color of gradient
   * @param endColorCode End color of gradient.
   * @param isVertical True if the gradient should be vertical.
   */
  void fillWithGradient(Rectangle rect, Color startColorCode, Color endColorCode, boolean isVertical);

  /**
   * @param polygon The polygon to fill with a gradient
   * @param startColorCode Start color of gradient
   * @param endColorCode End color of gradient.
   * @param isVertical True if the gradient should be vertical.
   */
  void fillWithGradient(Polygon polygon, Color startColorCode, Color endColorCode, boolean isVertical);
  
  /**
   * Fill a shape
   * 
   * @param shape The shape
   */
  void fillShape(Shape shape);

  /**
   * Draw a shape
   * 
   * @param shape The shape
   */
  void drawShape(Shape shape);

  /**
   * @param font The font
   * @return The font metrics
   */
  FontMetrics getFontMetrics(Font font);

  /**
   * Derive the style of the font
   * @param current Current font
   * @param style New style
   * @return The new font
   */
  Font deriveStyle(Font current, int style);
  
  /**
   * Derive the size of the font
   * @param current Current font
   * @param size New size
   * @return The new font
   */
  Font deriveSize(Font current, int size);
  
  /**
   * Draw an image object depending on the platform
   * @param image The image data
   * @param x image X
   * @param y image Y
   */
  void drawImage(Object image, int x, int y);
  
  /**
   * Draw an image object depending on the platform. 
   * The image is scaled to the width and height specified.
   * 
   * @param image The image data
   * @param x image X
   * @param y image Y
   * @param width image width
   * @param height image height
   */
  void drawImage(Object image, int x, int y, int width, int height);

  
  /**
   * Draws as much of the specified area of the specified image as is
   * currently available, scaling it on the fly to fit inside the
   * specified area of the destination drawable surface. Transparent pixels 
   * do not affect whatever pixels are already there.
   * <p>
   * @param       img the specified image to be drawn. This method does
   *                  nothing if <code>img</code> is null.
   * @param       dx1 the <i>x</i> coordinate of the first corner of the
   *                    destination rectangle.
   * @param       dy1 the <i>y</i> coordinate of the first corner of the
   *                    destination rectangle.
   * @param       dx2 the <i>x</i> coordinate of the second corner of the
   *                    destination rectangle.
   * @param       dy2 the <i>y</i> coordinate of the second corner of the
   *                    destination rectangle.
   * @param       sx1 the <i>x</i> coordinate of the first corner of the
   *                    source rectangle.
   * @param       sy1 the <i>y</i> coordinate of the first corner of the
   *                    source rectangle.
   * @param       sx2 the <i>x</i> coordinate of the second corner of the
   *                    source rectangle.
   * @param       sy2 the <i>y</i> coordinate of the second corner of the
   *                    source rectangle.
   * @since       oXygen 14.0
   */
  void drawImage(Object img,
            int dx1, int dy1, int dx2, int dy2,
            int sx1, int sy1, int sx2, int sy2);
  
  
  /**
   * Fills a rectangle with a pattern.
   * 
   * @param x The left corner x coordinate.
   * @param y The left corner y coordinate
   * @param width The width of the rectangle.
   * @param height The hight of the rectangle.
   * @param fillPattern One this class constants that begin with FILL_PATTERN_
   * @param fillColor The color for building the pattern.
   */
  void fillRect(int x, int y, int width, int height, int fillPattern, Color fillColor);

  /**
   * Draws a shadow using the shape of the polygon. The shadow will move to the right and down, with 
   * the size of the shadow.
   * 
   * @param shape The shape of the real object.
   * @param displacement The size of the shadow. This is the displacement to the original object, on X and Y.
   */
  void drawShadow(Shape shape, int displacement);
  
  /**
   * @param ellipse The ellipse to fill with a gradient
   * @param startColorCode Start color of gradient
   * @param endColorCode End color of gradient.
   * @param isVertical True if the gradient should be vertical.
   */
  void fillWithGradient(Ellipse ellipse, ro.sync.exml.view.graphics.Color startColorCode,
      ro.sync.exml.view.graphics.Color endColorCode, boolean isVertical);
  
  /**
   * @param shape The shape to fill with a gradient
   * @param startColorCode Start color of gradient
   * @param endColorCode End color of gradient.
   * @param isVertical True if the gradient should be vertical.
   */
  void fillWithGradient(Shape shape, ro.sync.exml.view.graphics.Color startColorCode,
      ro.sync.exml.view.graphics.Color endColorCode, boolean isVertical);
  
  /**
   * @param value The antialising value to be set depending on the platform.
   */
  void setAntialiasing(Object value);
  
  /**
   * @return The antialising value to be set depending on the platform.
   */
  Object getAntialiasing();
  
  /**
   * Activate the antialiasing mode.
   */
  void activateAntialiasing();
  
  /**
   * Disable the antialiasing mode.
   */
  void disableAntialiasing();

  /**
   * Draws a string at the specified coordinates. Takes into account the 
   * wrapping width and newline characters.
   * 
   * @param aStr The text, as an attributed string. 
   * @param x The x coordinate.
   * @param y The y coordinate.
   * @param wrappingWidth The maximum width of the wrapped text.
   */
  void drawWrappedString(AttributedString aStr, int x, int y, int wrappingWidth);

}