/*
 * The Syncro Soft SRL License
 * 
 * Copyright (c) 1998-2007 Syncro Soft SRL, Romania. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted
 * provided that the following conditions are met:
 * 
 * 1. Redistribution of source or in binary form is allowed only with the prior written permission
 * of Syncro Soft SRL.
 * 
 * 2. Redistributions of source code must retain the above copyright notice, this list of conditions
 * and the following disclaimer.
 * 
 * 3. Redistributions in binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other materials provided with
 * the distribution.
 * 
 * 4. The end-user documentation included with the redistribution, if any, must include the
 * following acknowledgment: "This product includes software developed by the Syncro Soft SRL
 * (http://www.sync.ro/)." Alternately, this acknowledgment may appear in the software itself, if
 * and wherever such third-party acknowledgments normally appear.
 * 
 * 5. The names "Oxygen" and "Syncro Soft SRL" must not be used to endorse or promote products
 * derived from this software without prior written permission. For written permission, please
 * contact contact@sync.ro.
 * 
 * 6. Products derived from this software may not be called "Oxygen", nor may "Oxygen" appear in
 * their name, without prior written permission of the Syncro Soft SRL.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SYNCRO SOFT SRL OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package ro.sync.exml.view.graphics;

import ro.sync.annotations.obfuscate.SkipObfuscate;
import ro.sync.annotations.obfuscate.SkipLevel;
import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;

/**
 * Rectangle.
 */
@API(type=APIType.EXTENDABLE, src=SourceType.PRIVATE)
@SkipObfuscate(classes = SkipLevel.PRIVATE, fields = SkipLevel.PRIVATE, methods = SkipLevel.PRIVATE)
public class Rectangle implements Shape {
  /**
   * The width.
   */
  public int width;

  /**
   * The height.
   */
  public int height;

  /**
   * The x.
   */
  public int x;

  /**
   * The y.
   */
  public int y;

  /**
   * Constructor.
   * 
   * @param x
   *          The x coordinate.
   * @param y
   *          The y coordinate.
   * @param width
   *          The width.
   * @param height
   *          The height.
   */
  public Rectangle(int x, int y, int width, int height) {
    this.width = width;
    this.height = height;
    this.x = x;
    this.y = y;
  }

  /**
   * Copy constructor.
   * 
   * @param r The other rectangle to be used.
   */
  public Rectangle(Rectangle r) {
    this.width = r.width;
    this.height = r.height;
    this.x = r.x;
    this.y = r.y;
  }

  /**
   * Determines whether or not this <code>Rectangle</code> and the specified
   * <code>Rectangle</code> intersect. Two rectangles intersect if their intersection is nonempty.
   * 
   * @param r
   *          the specified <code>Rectangle</code>
   * @return <code>true</code> if the specified <code>Rectangle</code> and this
   *         <code>Rectangle</code> intersect; <code>false</code> otherwise.
   */
  public final boolean intersects(Rectangle r) {
    int tw = this.width;
    int th = this.height;
    int rw = r.width;
    int rh = r.height;
    if (rw <= 0 || rh <= 0 || tw <= 0 || th <= 0) {
      return false;
    }
    int tx = this.x;
    int ty = this.y;
    int rx = r.x;
    int ry = r.y;
    rw += rx;
    rh += ry;
    tw += tx;
    th += ty;
    // overflow || intersect
    return ((rw < rx || rw > tx) && (rh < ry || rh > ty) && (tw < tx || tw > rx) && (th < ty || th > ry));
  }

  /**
   * Determines whether or not this <code>Rectangle</code> and the specified
   * <code>Rectangle</code> intersect. Two rectangles intersect if their intersection is nonempty.
   * @param rX X of rect
   * @param rY Y of rect
   * @param rW Width of rect
   * @param rH Height of rect
   * 
   * @return <code>true</code> if the specified <code>Rectangle</code> and this
   *         <code>Rectangle</code> intersect; <code>false</code> otherwise.
   */
  public final boolean intersects(int rX, int rY, int rW, int rH) {
    int tw = this.width;
    int th = this.height;
    if (rW <= 0 || rH <= 0 || tw <= 0 || th <= 0) {
      return false;
    }
    int tx = this.x;
    int ty = this.y;
    rW += rX;
    rH += rY;
    tw += tx;
    th += ty;
    // overflow || intersect
    return ((rW < rX || rW > tx) && (rH < rY || rH > ty) && (tw < tx || tw > rX) && (th < ty || th > rY));
  }

  /**
   * Checks whether or not this <code>Rectangle</code> entirely contains the specified
   * <code>Rectangle</code>.
   * 
   * @param r
   *          the specified <code>Rectangle</code>
   * @return <code>true</code> if the <code>Rectangle</code> is contained entirely inside this
   *         <code>Rectangle</code>;<code>false</code> otherwise.
   * @since JDK1.1
   */
  public boolean contains(Rectangle r) {
    return contains(r.x, r.y, r.width, r.height);
  }

  /**
   * Checks whether this <code>Rectangle</code> entirely contains the <code>Rectangle</code> at
   * the specified location ( <i>X </i>,&nbsp; <i>Y </i>) with the specified dimensions ( <i>W
   * </i>,&nbsp; <i>H </i>).
   * 
   * @param X
   *          the specified x coordinate
   * @param Y
   *          the specified y coordinate
   * @param W
   *          the width of the <code>Rectangle</code>
   * @param H
   *          the height of the <code>Rectangle</code>
   * @return <code>true</code> if the <code>Rectangle</code> specified by ( <i>X </i>,&nbsp;
   *         <i>Y </i>,&nbsp; <i>W </i>,&nbsp; <i>H </i>) is entirely enclosed inside this
   *         <code>Rectangle</code>;<code>false</code> otherwise.
   * @since JDK1.1
   */
  private boolean contains(int X, int Y, int W, int H) {
    int w = this.width;
    int h = this.height;
    if ((w | h | W | H) < 0) {
      // At least one of the dimensions is negative...
      return false;
    }
    // Note: if any dimension is zero, tests below must return false...
    int x = this.x;
    int y = this.y;
    if (X < x || Y < y) {
      return false;
    }
    w += x;
    W += X;
    if (W <= X) {
      // X+W overflowed or W was zero, return false if...
      // either original w or W was zero or
      // x+w did not overflow or
      // the overflowed x+w is smaller than the overflowed X+W
      if (w >= x || W > w) {
        return false;
      }
    } else {
      // X+W did not overflow and W was not zero, return false if...
      // original w was zero or
      // x+w did not overflow and x+w is smaller than X+W
      if (w >= x && W > w) {
        return false;
      }
    }
    h += y;
    H += Y;
    if (H <= Y) {
      if (h >= y || H > h) {
        return false;
      }
    } else {
      if (h >= y && H > h) {
        return false;
      }
    }
    return true;
  }

  /**
   * Checks whether or not this <code>Rectangle</code> contains the 
   * point at the specified location
   * (<i>x</i>,&nbsp;<i>y</i>).
   * @param  x the specified x coordinate
   * @param  y the specified y coordinate
   * @return    <code>true</code> if the point 
   *            (<i>x</i>,&nbsp;<i>y</i>) is inside this 
   *      <code>Rectangle</code>; 
   *            <code>false</code> otherwise.
   * @since     JDK1.1
   */
  public boolean contains(int x, int y) {
    return inside(x, y);
  }
  
  /**
   * Checks whether or not this <code>Rectangle</code> contains the 
   * point at the specified location
   * (<i>X</i>,&nbsp;<i>Y</i>).
   * @param  X the specified x coordinate
   * @param  Y the specified y coordinate
   * @return    <code>true</code> if the point 
   *            (<i>X</i>,&nbsp;<i>Y</i>) is inside this 
   *      <code>Rectangle</code>; 
   *            <code>false</code> otherwise.
   * @deprecated As of JDK version 1.1,
   * replaced by <code>contains(int, int)</code>.
   */
  @Deprecated
  private boolean inside(int X, int Y) {
    int w = this.width;
    int h = this.height;
    if ((w | h) < 0) {
      // At least one of the dimensions is negative...
      return false;
    }
//  Note: if either dimension is zero, tests below must return false...
    int x = this.x;
    int y = this.y;
    if (X < x || Y < y) {
      return false;
    }
    w += x;
    h += y;
//  overflow || intersect
    return ((w < x || w > X) &&
        (h < y || h > Y));
  }
  
  /**
   * Returns a <code>String</code> representing this 
   * <code>Rectangle</code> and its values.
   * @return a <code>String</code> representing this 
   *               <code>Rectangle</code> object's coordinate and size values.
   */
  @Override
  public String toString() {
    return "[x=" + x + ",y=" + y + ",width=" + width + ",height=" + height + "]";
  }
  
  /**
   * Checks whether two rectangles are equal.
   * <p>
   * The result is <code>true</code> if and only if the argument is not 
   * <code>null</code> and is a <code>Rectangle</code> object that has the 
   * same top-left corner, width, and height as this <code>Rectangle</code>. 
   * @param obj the <code>Object</code> to compare with
   *                this <code>Rectangle</code>
   * @return    <code>true</code> if the objects are equal; 
   *            <code>false</code> otherwise.
   */
  @Override
  public boolean equals(Object obj) {
    if (obj instanceof Rectangle) {
      Rectangle r = (Rectangle)obj;
      return ((x == r.x) &&
          (y == r.y) &&
          (width == r.width) &&
          (height == r.height));
    }
    return false;
  }
  
  /**
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    return x ^ y + width ^ height;
  }
  
  /**
   * Computes the union of this <code>Rectangle</code> with the 
   * specified <code>Rectangle</code>. Returns a new 
   * <code>Rectangle</code> that 
   * represents the union of the two rectangles
   * @param r the specified <code>Rectangle</code>
   * @return    the smallest <code>Rectangle</code> containing both 
   *      the specified <code>Rectangle</code> and this 
   *      <code>Rectangle</code>.
   */
  public Rectangle union(Rectangle r) {
    int x1 = Math.min(x, r.x);
    int x2 = Math.max(x + width, r.x + r.width);
    int y1 = Math.min(y, r.y);
    int y2 = Math.max(y + height, r.y + r.height);
    return new Rectangle(x1, y1, x2 - x1, y2 - y1);
  }

  /**
   * @return  The rectangle height.
   */
  public int getHeight() {
    return height;
  }
  
  /**
   * @return  The rectangle width.
   */
  public int getWidth() {
    return width;
  }
  
  /**
   * @return  The rectangle x coordinate.
   */
  public int getX() {
    return x;
  }
  
  /**
   * @return  The rectangle y coordinate.
   */
  public int getY() {
    return y;
  }
  
  /**
   * @param r The rectangle to check for adjacency.
   * @return True if the rectangles are neighbours around one margin
   * and they do not intersect each other.
   */
  public boolean adjacent(Rectangle r) {
    boolean adjacent = false;
    if (!intersects(r)) {
      if (new Rectangle(x - 1, y - 1, width + 2, height + 2).intersects(r)) {
        adjacent = true;
      }
    }    
    return adjacent;
  }

  /**
   * Computes the intersection of this <code>Rectangle</code> with the 
   * specified <code>Rectangle</code>. Returns a new <code>Rectangle</code> 
   * that represents the intersection of the two rectangles.
   * If the two rectangles do not intersect, the result will be
   * an empty rectangle.
   *
   * @param     r   the specified <code>Rectangle</code>
   * @return    the largest <code>Rectangle</code> contained in both the 
   *            specified <code>Rectangle</code> and in 
   *      this <code>Rectangle</code>; or if the rectangles
   *            do not intersect, an empty rectangle.
   */
  public Rectangle intersection(Rectangle r) {
    int tx1 = this.x;
    int ty1 = this.y;
    int rx1 = r.x;
    int ry1 = r.y;
    long tx2 = tx1; tx2 += this.width;
    long ty2 = ty1; ty2 += this.height;
    long rx2 = rx1; rx2 += r.width;
    long ry2 = ry1; ry2 += r.height;
    if (tx1 < rx1) {
      tx1 = rx1;
    }
    if (ty1 < ry1) {
      ty1 = ry1;
    }
    if (tx2 > rx2) {
      tx2 = rx2;
    }
    if (ty2 > ry2) {
      ty2 = ry2;
    }
    tx2 -= tx1;
    ty2 -= ty1;
    //  tx2,ty2 will never overflow (they will never be
    //  larger than the smallest of the two source w,h)
    //  they might underflow, though...
    if (tx2 < Integer.MIN_VALUE) {
      tx2 = Integer.MIN_VALUE;
    }
    if (ty2 < Integer.MIN_VALUE) {
      ty2 = Integer.MIN_VALUE;
    }
    return new Rectangle(tx1, ty1, (int) tx2, (int) ty2);
  }
  
  /**
   * TC main.
   * @param args Main args
   */
  public static void main(String[] args) {
    Rectangle r1 = new Rectangle(0, 0, 2, 2);
    Rectangle r2 = new Rectangle(2, 2, 2, 2);
    System.out.println(r1.intersects(r2));
  }

  /**
   * @see ro.sync.exml.view.graphics.Shape#getBounds()
   */
  public Rectangle getBounds() {
    return this;
  }

  /**
   * @see ro.sync.exml.view.graphics.Shape#translate(int, int)
   */
  public Shape translate(int tx, int ty) {
    return new Rectangle(x + tx, y + ty, width, height);
  }
  
  /**
   * @see java.lang.Object#clone()
   */
  @Override
  public Object clone() {
    return new Rectangle(x, y, width, height);
  }
}