/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.exml.workspace.api.editor.page;

import javax.swing.text.BadLocationException;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.document.DocumentPositionedInfo;
import ro.sync.exml.view.graphics.Point;
import ro.sync.exml.view.graphics.Rectangle;

/**
 * Provides access to methods related to the editor actions and information for 
 * the Text and Author pages.
 * 
 * @since 11.2 
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PUBLIC)
public interface WSTextBasedEditorPage extends WSEditorPage {
  /**
   * Get selection start offset. It is inclusive.
   * 
   * @return The selection start offset, 0 based.
   */
  int getSelectionStart();
  
  /**
   * Get selection end offset. It is exclusive.
   * 
   * @return The selection end offset, 0 based.
   */
  int getSelectionEnd();
  
  /**
   * Get the selected text. The text does not contain XML tags for the Author page. 
   * 
   * @return The selected text or an empty string if no selection is present.
   */
  String getSelectedText();
  
  /**
   * The current caret offset.
   * 
   * @return The caret offset, 0 based.
   */
  int getCaretOffset();
  
  /**
   * Delete the selected text, if any.
   */
  void deleteSelection();
  
  /**
   * Check if the editor page has a selection
   * 
   * @return <code>true</code> if there is a selection, <code>false</code> otherwise.
   */
  boolean hasSelection();
  
  /**
   * Select the word at caret, if any.
   */
  void selectWord();
  
  /**
   * Move the caret to the specified offset.
   * 
   * @param offset The offset where the caret should be positioned, 0 based.
   */
  void setCaretPosition(int offset);
  
  /**
   * Select the interval between start and end offset.
   *  
   * @param startOffset Inclusive start offset 
   * @param endOffset  Exclusive end offset
   */
  void select(int startOffset, int endOffset);

  /**
   * Compute the offsets of the word that contains the caret position.
   * 
   * @return An array with the start and end offsets of the word at caret or
   * <code>null</code> if the values couldn't be obtained.
   * The start offset is inclusive and greater or equal than 1.
   * The end offset is also inclusive and less or equal than the content length.
   */
  int[] getWordAtCaret();
  
  /**
   * Take relative mouse coordinates and translate then to absolute on-screen coordinates.
   * 
   * @param x The "x" coordinate relative to the viewport origin.
   * @param y The "y" coordinate relative to the viewport origin.
   * @return  A point with the "x" and "y" coordinates relative to the screen. 
   *          It does not return a <code>null</code> Point.
   */
  Point getLocationOnScreenAsPoint(int x, int y);
  
  /**
   * Take mouse coordinates relative to the main application frame/display and translate then to coordinates relative to the editing area.
   * 
   * @param x The "x" coordinate which is relative to the main application frame/display.
   * @param y The "y" coordinate which is relative to the main application frame/display.
   * @return  A point with the "x" and "y" coordinates relative to the editing component area. 
   *          It does not return a <code>null</code> Point.
   *          
   * For example an Eclipse org.eclipse.swt.dnd.DropTargetEvent has the x and y coordinates relative to the Display.
   * In order to use API like viewToModelOffset(), the coordinates need to be translated relative to the editing area.
   * 
   *  @since 14.2
   *  
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one of the next versions of the application. If you have suggestions, 
   * comments about it, please let us know.</p> 
   */
  Point getLocationRelativeToEditorFromScreen(int x, int y);
  
  /**
   * Returns a representation of the caret shape for the specified document offset.
   * 
   * @param offset The document offset to get the corresponding caret shape for. 
   * @return The caret rectangle shape. It does not return a <code>null</code> Rectangle.
   *  
   */
  Rectangle modelToViewRectangle(int offset);
  
  /**
   * Get the position in the document corresponding to the point in the editing component.
   * 
   * @param x The "x" coordinate relative to the editing component origin.
   * @param y The "y" coordinate relative to the editing component origin.
   * @return  The offset corresponding to the given point. The method does not return <code>null</code>, 
   * instead an undefined view to model info object is returned if a valid one could not be determined.
   */
  int viewToModelOffset(int x, int y);
  
  /**
   * Sets the specified flag to indicate whether or not this page should be editable.
   *
   * @param editable <code>true</code> if the page should be editable.
   * 
   * @since 12.2
   */
  public void setEditable(boolean editable);
  
  /**
   * Check if the document can be edited. A document can be set as read-only 
   * either in the file system or from API, by using the 
   * {@link #setEditable(boolean)} method.
   * 
   * @return <code>true</code> if the document is editable.
   * 
   * @since 14.1
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one
   * of the next versions of the application. If you have suggestions, 
   * comments about it, please let us know.</p>
   */
  public boolean isEditable();
  
  /**
   * Find the start and end offsets that correspond to a document position information
   * {@link DocumentPositionedInfo}. 
   * 
   * @param dpInfo The document position information.
   * @return an array of length two containing the start offset at index 0
   * and the end offset at index 1.
   * 
   * @throws BadLocationException If the document position info does not correspond to a 
   * valid position in document.
   * 
   * @since 14.1
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one
   * of the next versions of the application. If you have suggestions, 
   * comments about it, please let us know.</p>
   */
  public int[] getStartEndOffsets(final DocumentPositionedInfo dpInfo) throws BadLocationException;
}