/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.exml.workspace.api.editor.page.text;

import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.exml.workspace.api.editor.page.WSTextBasedEditorPage;
import ro.sync.exml.workspace.api.editor.page.text.actions.TextActionsProvider;

/**
 * Text editor page access. 
 * 
 * @since 11.2 
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PUBLIC)
public interface WSTextEditorPage extends WSTextBasedEditorPage {

  /**
   * Get the edited document. For eclipse, the returned instance is an javax.swing.text.Document adapter over the Eclipse native document.
   * 
   * @return The edited document.
   */
  Document getDocument();
  
  /**
   * Get the internal text component which is currently used for editing.
   * 
   * @return for the stand alone version, a javax.swing.JTextArea and for the eclipse implementation a org.eclipse.swt.custom.StyledText.
   * 
   * @since 12
   */
  Object getTextComponent();
  
  /**
   * Get the schema manager used to ask useful information about allowed elements and the context of the current offset.
   * @return the XML Schema Manager used to ask useful information about allowed elements and the context of the current offset.
   * Only available for opened XML files. 
   * 
   * @since 12.1
   */
  WSTextXMLSchemaManager getXMLSchemaManager();
  
  /**
   * Begin a compound undoable edit operation. This is useful if you make modifications through the API and want Oxygen to undo in a single step.
   * 
   * This should be used like:
   * <pre>try{
   *  beginCompoundUndoableEdit();
   *  //YOUR CODE HERE
   * } finally{
   *  endCompoundUndoableEdit();
   * }
   * </pre>
   * 
   * @since 12.2
   */
  void beginCompoundUndoableEdit();
  
  /**
   * End a compound undoable edit operation. This is useful if you make modifications through the API and want Oxygen to undo in a single step.
   * 
   * This should be used like:
   * <pre>try{
   *  beginCompoundUndoableEdit();
   *  //YOUR CODE HERE
   * } finally{
   *  endCompoundUndoableEdit();
   * }
   * </pre>
   * @since 12.2
   */
  void endCompoundUndoableEdit();
  
  /**
   *  Converts an offset position (a position between the characters of the
   *  text) to the line number that contains the offset. If the offset is
   *  between two lines, then the number of the line after is returned. The
   *  line number returned is indexed in 1.
   *
   * @param  offset                 Offset in document.
   *
   * @return                        The line which contains the specified
   *      offset. 1 based.
   * @throws  BadLocationException  When the offset is not valid.
   * @since 14
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one of the next versions of the application. If you have suggestions,
   * comments about it, please let us know.</p>
   */
  int getLineOfOffset(int offset) throws BadLocationException;
  
  /**
   *  Converts an offset position (a position between the characters of the
   *  text) to the column number in the line that contains the offset.
   *
   * @param  offset                 The offset that is to be checked.
   *
   * @return                        The column in the line that contains the
   *      offset.
   * @throws  BadLocationException  Bad Location Exception.
   * @since 14
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one of the next versions of the application. If you have suggestions,
   * comments about it, please let us know.</p>
   */
  int getColumnOfOffset(int offset) throws BadLocationException;
  
  /**
   *  Gets the offset of the start of the specified line (return >=0). The
   *  line number is indexed in 1.
   *
   * @param  lineNumber                 The number of the line. Indexed in 1.
   *
   * @return                            The offset of the start of the
   *      specified line.
   * @throws BadLocationException  When line does not exist.
   * @throws  IllegalArgumentException  If the line number is out of range.
   * @since 14
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one of the next versions of the application. If you have suggestions,
   * comments about it, please let us know.</p>
   */
  int getOffsetOfLineStart(int lineNumber) throws BadLocationException;

  /**
   *  Gets the offset of the end of the specified line (return >=0). This will
   *  be equal to the start offset of the next line, if there is one. The line
   *  number is indexed in 1.
   *
   * @param  lineNumber                 The number of the line. Indexed in 1.
   *
   * @return                            The offset of the end of the specified
   *      line.
   * @throws  BadLocationException  If the line number is out of range.
   * @since 14
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one of the next versions of the application. If you have suggestions,
   * comments about it, please let us know.</p>
   */
  public int getOffsetOfLineEnd(int lineNumber) throws BadLocationException;
  
  /**
   * Add the pop-up menu customizer which can be used to customize the pop-up menu (add/remove actions) before showing it in the Text page.
   * If the customizer is already added, it will not be added again.
   * 
   * @param popUpCustomizer the pop-up menu customizer.
   * @since 14.1
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one of the next versions of the application. If you have suggestions, 
   * comments about it, please let us know.</p>   
   */
  void addPopUpMenuCustomizer(TextPopupMenuCustomizer popUpCustomizer);
  
  /**
   * Remove the pop-up menu customizer which is used to customize the pop-up menu (add/remove actions) before showing it in the Text page.
   * 
   * @param popUpCustomizer the pop-up menu customizer.
   * @since 14.1
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one of the next versions of the application. If you have suggestions, 
   * comments about it, please let us know.</p>   
   */
  void removePopUpMenuCustomizer(TextPopupMenuCustomizer popUpCustomizer);
  
  /**
   * Provides access to actions already defined in the Text page like: Undo, Redo, etc.
   * @return access to actions already defined in the Text page.
   * 
   * @since 14.2
   * <br>
   * <br>
   * *********************************
   * <br>
   * EXPERIMENTAL - Subject to change
   * <br>
   * ********************************
   * </br>
   * <p>Please note that this API is not marked as final and it can change in one
   * of the next versions of the application. If you have suggestions, 
   * comments about it, please let us know.</p> 
   */
  TextActionsProvider getActionsProvider();
}
