/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.util.editorvars;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.annotations.obfuscate.SkipLevel;
import ro.sync.annotations.obfuscate.SkipObfuscate;
import ro.sync.exml.OxygenSystemProperties;
import ro.sync.exml.options.OptionTags;
import ro.sync.exml.options.Options;
import ro.sync.io.ExceptCVSAndSVNFileFilter;
import ro.sync.io.FileSystemUtil;
import ro.sync.util.ArchiveExtensionsRecognizer;
import ro.sync.util.ArchiveURLProvider;
import ro.sync.util.Formatter;
import ro.sync.util.OxygenDirectoryUtil;
import ro.sync.util.PlatformDetector;
import ro.sync.util.PreferencesDirectoryDetector;
import ro.sync.util.ProjectURLProvider;
import ro.sync.util.URLUtil;
import ro.sync.util.Util;

/**
 * Holds constants representing all editor variables defined in Oxygen.
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PRIVATE)
@SkipObfuscate(classes = SkipLevel.PRIVATE, fields = SkipLevel.PRIVATE, methods = SkipLevel.PRIVATE)
public final class EditorVariables {
  
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(EditorVariables.class.getName());
  /**
   * Used to determine how framework variables should be expanded/rewritten. 
   */
  public static enum FrameworkRewritePolicy {
    /**
     * All framework variables will be expanded to their absolute paths.
     */
    REWRITE_ABSOLUTE, 
    /**
     * Variables FRAMEWORKS_DIRECTORY and FRAMEWORKS_DIRECTORY_URL will not be expanded
     * and will be used to rewrite FRAMEWORK_DIRECTORY and FRAMEWORK_URL if encountered
     * and if they match. If they don't match, FRAMEWORK_DIRECTORY and FRAMEWORK_URL
     * will be expanded to their absolute paths.
     */
    REWRITE_WITH_FRAMEWORKS,
    /**
     * Variables FRAMEWORK_DIRECTORY and FRAMEWORK_URL will not be expanded
     * and will be used to rewrite FRAMEWORKS_DIRECTORY and FRAMEWORKS_DIRECTORY_URL 
     * if they match. If they don't match, FRAMEWORKS_DIRECTORY and FRAMEWORKS_DIRECTORY_URL 
     * will be expanded to their absolute path.
     */
    REWRITE_WITH_FRAMEWORK
  }
  /**
   * The editor variables prefix
   */
  public static final String EDITOR_VARIABLES_PREFIX = "${";
  /**
   * The current file directory url editor variable.
   * The value is <code>${cfdu}</code>. 
   */
  public static final String CURRENT_FILE_DIRECTORY_URL = "${cfdu}";
  /**
   * The current file directory editor variable.
   * The value is <code>${cfd}</code>. 
   */
  public static final String CURRENT_FILE_DIRECTORY = "${cfd}";
  /**
   * The current file url editor variable.
   * The value is <code>${currentFileURL}</code>. 
   */
  public static final String CURRENT_FILE_URL = "${currentFileURL}";
  /**
   * The current file url editor variable old version.
   * The value is <code>${cfu}</code>. 
   */
  public static final String CURRENT_FILE_URL_OLD = "${cfu}";
  /**
   * The current file editor variable.
   * The value is <code>${cf}</code>. 
   */
  public static final String CURRENT_FILE = "${cf}";
  /**
   * The current filename editor variable.
   * The value is <code>${cfn}</code>. 
   */
  public static final String CURRENT_FILENAME = "${cfn}";
  /**
   * The current filename extension editor variable.
   * The value is <code>${cfne}</code>. 
   */
  public static final String CURRENT_FILENAME_WITH_EXTENSION = "${cfne}";
  /**
   * The project directory URL editor variable.
   * The value is <code>${pdu}</code>. 
   */
  public static final String PROJECT_DIRECTORY_URL = "${pdu}";
  /**
   * The project directory editor variable.
   * The value is <code>${pd}</code>. 
   */
  public static final String PROJECT_DIRECTORY = "${pd}";
  /**
   * The project name editor variable
   * The value is <code>${pn}</code>.
   */
  public static final String PROJECT_NAME = "${pn}";

  /**
   * The archive name editor variable.
   */
  public static final String ARCHIVE_NAME = "${afn}";
  
  /**
   * The archive name with extension editor variable.
   */
  public static final String ARCHIVE_NAME_WITH_EXTENSION = "${afne}";
  
  /**
   * The archive path editor variable.
   */
  public static final String ARCHIVE_PATH = "${af}";
  
  /**
   * The archive path URL editor variable.
   */
  public static final String ARCHIVE_PATH_URL = "${afu}";
  
  /**
   * The archive directory editor variable.
   */
  public static final String ARCHIVE_FILE_DIRECTORY = "${afd}";
  
  /**
   * The archive directory URL editor variable.
   */
  public static final String ARCHIVE_FILE_DIRECTORY_URL = "${afdu}";
  
  /**
   * The frameworks directory URL editor variable.
   * The value is <code>${frameworks}</code>. 
   */
  public static final String FRAMEWORKS_DIRECTORY_URL = "${frameworks}";
  /**
   * The user home URL editor variable.
   * The value is <code>${home}</code>. 
   */
  public static final String USER_HOME_URL = "${home}";
  /**
   * The frameworks directory editor variable.
   * The value is <code>${frameworksDir}</code>. 
   */
  public static final String FRAMEWORKS_DIRECTORY = "${frameworksDir}";
  /**
   * Represents the framework directory where the document type that contains 
   * this reference resides.
   * 
   * The value is <code>${frameworkDir}</code>. 
   */
  public static final String FRAMEWORK_DIRECTORY = "${frameworkDir}";
  /**
   * The framework directory URL editor variable.
   * The value is <code>${framework}</code>. 
   */
  public static final String FRAMEWORK_URL = "${framework}";
  /**
   * The user home dir editor variable.
   * The value is <code>${homeDir}</code>. 
   */
  public static final String USER_HOME_DIR = "${homeDir}";
  /**
   * The oxygen install dir editor variable.
   * The value is <code>${oxygenInstallDir}</code>. 
   */
  public static final String OXYGEN_INSTALL_DIR = "${oxygenInstallDir}";
  /**
   * The detected schema editor variable.
   * The value is <code>${ds}</code>.
   */
  public static final String DETECTED_SCHEMA = "${ds}";
  /**
   * The detected schema URL editor variable.
   * The value is <code>${dsu}</code>. 
   */
  public static final String DETECTED_SCHEMA_URL = "${dsu}";
  /**
   * The transformation saved file editor Variable.
   * The value is <code>${tsf}</code>. 
   */
  public static final String TRANSFORMATION_SAVED_FILE = "${tsf}";
  /**
   * The debugger XML input editor variable.
   * The value is <code>${dbgXML}</code>. 
   */
  public static final String DEBUGGER_XML_SOURCE = "${dbgXML}";
  /**
   * The debugger XSL input editor variable.
   * The value is <code>${dbgXSL}</code>. 
   */
  public static final String DEBUGGER_XSL_SOURCE = "${dbgXSL}";
  /**
   * The FOP transformation method editor variable.
   * The value is <code>${method}</code>. 
   */
  public static final String FOP_TRANSFORMATION_METHOD = "${method}";
  /**
   * The FOP Antenna House transformation method editor variable.
   * The value is <code>${ah_method}</code>. 
   */
  public static final String FOP_AH_TRANSFORMATION_METHOD = "${ah_method}";
  /**
   * The FO input file editor variable.
   * The value is <code>${fo}</code>. 
   */
  public static final String FO_INPUT_FILE = "${fo}";
  /**
   * The output file editor variable.
   * The value is <code>${out}</code>. 
   */
  public static final String OUTPUT_FILE = "${out}";
  /**
   * The output file URL editor variable.
   * The value is <code>${outu}</code>. 
   */
  public static final String OUTPUT_FILE_URL = "${outu}";  
  /**
   * The Oxygen Home URL editor variable.
   * The value is <code>${oxygenHome}</code>.
   */
  public static final String OXYGEN_HOME_URL = "${oxygenHome}";

  /**
   * XML Document editor variable.
   * The value is <code>${xml}</code>.
   */
  public static final String XML = "${xml}";
  
  /**
   * XSL Document editor variable
   * The value is <code>${xsl}</code>.
   */
  public static final String XSL = "${xsl}";
  
  /**
   * SQL Document editor variable
   * The value is <code>${sql}</code>.
   */
  public static final String SQL = "${sql}";
  
  /**
   * XQuery Document editor variable
   * The value is <code>${xquery}</code>.
   */
  public static final String XQUERY = "${xq}";
  
  /**
   * XProc Document editor variable
   * The value is <code>${xproc}</code>.
   */
  public static final String XPROC = "${xproc}";
  
  /**
   * XML Document URL editor variable
   * The value is <code>${xmlu}</code>.
   */
  public static final String XML_URL = "${xmlu}";
  
  /**
   * XSL Document URL editor variable
   * The value is <code>${xslu}</code>.
   */
  public static final String XSL_URL = "${xslu}";
  
  /**
   * XSL Document URL editor variable
   * The value is <code>${xslu}</code>.
   */
  public static final String SQL_URL = "${sqlu}";
  
  /**
   * XSL Document URL editor variable
   * The value is <code>${xslu}</code>.
   */
  public static final String XQUERY_URL = "${xqu}";
  
  /**
   * XSL Document URL editor variable
   * The value is <code>${xslu}</code>.
   */
  public static final String XPROC_URL = "${xprocu}";
  
  /**
   * Path separator editor variable. Expands to ":" on Linux and Mac, ";" on Windows
   * The value is <code>${ps}</code>.
   */
  public static final String PATH_SEPARATOR = "${ps}";
  
  ////Some SVN editor variables.
  /**
   * The first file to diff.
   * The value is <code>${firstFile}</code>.
   */
  public static final String FIRST_FILE_TO_DIFF = "${firstFile}";
  
  /**
   * The second file to diff.
   * The value is <code>${secondFile}</code>.
   */
  public static final String SECOND_FILE_TO_DIFF = "${secondFile}";
  /**
   * The ancestor file to diff.
   * The value is <code>${ancestorFile}</code>.
   */
  public static final String ANCESTOR_FILE_TO_DIFF = "${ancestorFile}";
  
  /**
   * The time stamp.
   * The value is <code>${timeStamp}</code>.
   */
  public static final String TIMESTAMP = "${timeStamp}";
  
  /**
   * The UUID.
   * The value is <code>${uuid}</code>.
   */
  public static final String UUID = "${uuid}";

  /**
   * The unique friendly ID generated by Oxygen.
   * The value is <code>${id}</code>.
   */
  public static final String ID = "${id}";
  
  /**
   * Used to show the user the options he has to ask for a value.
   */
  public static final String ASK_PARAM_VALUE_TEMPLATE = "${ask('Message', input_type, 'default_value')}";
  
  /**
   * Prefix for environment functions
   */
  static final String ENV_FUNCTION_VARIABLE_PREFIX = "${env(";
  /**
   * Prefix for system functions
   */
  static final String SYSTEM_FUNCTION_VARIABLE_PREFIX = "${system(";
  /**
   * Prefix for translate functions
   */
  public static final String TRANSLATE_FUNCTION_VARIABLE_PREFIX = "${i18n(";
  /**
   * Prefix for date functions
   */
  static final String DATE_FUNCTION_VARIABLE_PREFIX = "${date(";
  /**
   * Suffix for functions (env or system)
   */
  public static final String FUNCTION_VARIABLE_SUFFIX = ")}";
  
  /**
   * ENV var name
   */
  static final String ENV_VAR_NAME = "VAR_NAME";
  
  /**
   * SYSTEM var name
   */
  static final String SYSTEM_VAR_NAME = "var.name";
  
  /**
   * Sample of environment function specification
   */
  static final String ENV_FUNCTION_SAMPLE = ENV_FUNCTION_VARIABLE_PREFIX +  ENV_VAR_NAME + FUNCTION_VARIABLE_SUFFIX;
  
  /**
   * Sample of system function specification
   */
  static final String SYSTEM_FUNCTION_SAMPLE = SYSTEM_FUNCTION_VARIABLE_PREFIX + SYSTEM_VAR_NAME + FUNCTION_VARIABLE_SUFFIX;
  
  /**
   * DATE sample format
   */
  static final String DATE_FORMAT_SAMPLE = "yyyy-MM-dd";
  
  /**
   * Sample of date function specification
   */
  public static final String DATE_FUNCTION_SAMPLE = DATE_FUNCTION_VARIABLE_PREFIX + DATE_FORMAT_SAMPLE + FUNCTION_VARIABLE_SUFFIX;
  
  /**
   * The caret editor variable.
   */
  public static final String CT_CARET_EDITOR_VARIABLE = "${caret}";
  /**
   * The caret editor variable.
   */
  public static final String CT_SELECTION_EDITOR_VARIABLE = "${selection}";
  
  /**
   * Unique marker used to position the caret
   */
  public static final String UNIQUE_CARET_MARKER_PI_NAME_FOR_AUTHOR = "UNIQUE_CARET_MARKER_FOR_AUTHOR";
  
  /**
   * Unique marker used to position the caret
   */
  public static final String UNIQUE_CARET_MARKER_FOR_AUTHOR = "<?" + UNIQUE_CARET_MARKER_PI_NAME_FOR_AUTHOR+"?>";
  
  
  /**
   * <code>true</code> if the frameworks directory was initialized from options, and must use
   * frameworks directory from the system properties. Otherwise it will be taken from options.
   */
  static boolean frameworksDirInitialized = false;
  
  /**
   * All known editor variables set. This has to be declared before the EXML messages.
   * EXML messages use expandEditorVariables() which needs the set instantiated. 
   */
  static final Set EDITOR_VARIABLES_SET = new HashSet() {
    {
      // Init the editor variables to descriptions map.
      add(CURRENT_FILE_DIRECTORY);
      add(CURRENT_FILE_DIRECTORY_URL);
      add(CURRENT_FILE);
      add(CURRENT_FILE_URL);
      add(CURRENT_FILE_URL_OLD);
      add(CURRENT_FILENAME);
      add(CURRENT_FILENAME_WITH_EXTENSION);
      add(PROJECT_DIRECTORY_URL);
      add(PROJECT_DIRECTORY);
      add(PROJECT_NAME);
      add(FRAMEWORKS_DIRECTORY_URL);
      add(USER_HOME_URL);
      add(FRAMEWORKS_DIRECTORY);
      add(USER_HOME_DIR);
      add(OXYGEN_INSTALL_DIR);
      add(OXYGEN_HOME_URL);
      add(DETECTED_SCHEMA);
      add(DETECTED_SCHEMA_URL);
      add(TRANSFORMATION_SAVED_FILE);
      add(DEBUGGER_XML_SOURCE);
      add(DEBUGGER_XSL_SOURCE);
      add(FOP_TRANSFORMATION_METHOD);
      add(FOP_AH_TRANSFORMATION_METHOD);
      add(FO_INPUT_FILE);
      add(OUTPUT_FILE);
      add(OUTPUT_FILE_URL);
      add(XML);
      add(XSL);
      add(SQL);
      add(XQUERY);
      add(XPROC);
      add(XML_URL);
      add(XSL_URL);
      add(SQL_URL);
      add(XQUERY_URL);
      add(XPROC_URL);
      add(PATH_SEPARATOR);
      add(TIMESTAMP);
      add(UUID);
      add(ID);
      add(UNIQUE_CARET_MARKER_FOR_AUTHOR);
      add(ARCHIVE_NAME);
      add(ARCHIVE_NAME_WITH_EXTENSION);
      add(ARCHIVE_PATH);
      add(ARCHIVE_PATH_URL);
      add(ARCHIVE_FILE_DIRECTORY);
      add(ARCHIVE_FILE_DIRECTORY_URL);
      add(FRAMEWORK_DIRECTORY);
      add(FRAMEWORK_URL);
    }
  };
  
  /**
   * Provides access to the current loaded project URL.
   */
  private static ProjectURLProvider projectURLProvider;
  
  /**
   * Provides access to the current archive opened in the Archive Browser.
   */
  private static ArchiveURLProvider archiveURLProvider;
  
  /**
   * Provides access to the supported archive extensions.
   */
  private static ArchiveExtensionsRecognizer archiveExtensionsRecognizer;
  
  /**
   * Returns a description of the editor variable.
   * 
   * @param editorVariable The editor variable to get description for.
   * @return The description for the editor variable.
   */
  public static String getDescription(String editorVariable) {
    String descr =
      (String) EditorVariablesToDescriptions.EDITOR_VARIABLE_TO_DESCRIPTION_MAP.get(editorVariable);
    if (descr == null) {
      //Maybe user defined variable
      UserEditorVariablePO[] userEditorVars =
        (UserEditorVariablePO[]) Options.getInstance().getObjectArrayProperty(
            OptionTags.USER_DEFINED_EDITOR_VARIABLES, UserEditorVariablePO.class);
      for (int i = 0; i < userEditorVars.length; i++) {
        String varName = userEditorVars[i].getVarName();
        if (varName.equals(editorVariable)) {
          descr = userEditorVars[i].getVarShortDescription();
          break;
        }
      }
    }
    return descr;
  }
  
  /**
   * Checks if the given expression contains editor variables.
   * 
   * @param expression The expression to check.
   *
   * @return <code>true</code> if the expression contains one of the available editor variables.
   */
  public static boolean containsEditorVariable(String expression) {
    boolean toReturn = false;
    if (expression != null) {
      Iterator<String> iter = EDITOR_VARIABLES_SET.iterator();
      while (iter.hasNext()) {
        String editorVariable = iter.next();
        if (expression.contains(editorVariable)) {
          toReturn = true;
          break;
        }
      }
      if (!toReturn) {
        // Maybe user defined editor variable
        UserEditorVariablePO[] userEditorVars =
          (UserEditorVariablePO[]) Options.getInstance().getObjectArrayProperty(
              OptionTags.USER_DEFINED_EDITOR_VARIABLES, UserEditorVariablePO.class);
        for (int i = 0; i < userEditorVars.length; i++) {
          String varName = userEditorVars[i].getVarName();
          if (expression.contains(varName)) {
            toReturn = true;
            break;
          }
        }
      }
      if (!toReturn) {
        // Maybe it contains some environment or system functions which can be expanded
        if (expression.contains(ENV_FUNCTION_VARIABLE_PREFIX) 
            || expression.contains(SYSTEM_FUNCTION_VARIABLE_PREFIX)
            || expression.contains(DATE_FUNCTION_VARIABLE_PREFIX)) {
          if (expression.contains(FUNCTION_VARIABLE_SUFFIX)) {
            toReturn = true;
          }
        }
      }
    }
    return toReturn;
  }
  
  /**
   * Expand the editor variables in the output file name.
   * The currently known editor variables are declared in this class. 
   * 
   * @param expr                  The expresion containing editor variables.
   * @param currentEditedFileURL  The full path of the current edited file, as an URI.
   *
   * @return The expresion with the editor variables expanded, possibly an URI.
   */
  public static String expandEditorVariables(String expr, String currentEditedFileURL) {
    if (!containsEditorVariable(expr)) {
      return expr;
    }
    
    //Expand user defined editor variables first so that if they contain other Oxygen editor variables, 
    //they will also be expanded.
    UserEditorVariablePO[] userEditorVars =
      (UserEditorVariablePO[]) Options.getInstance().getObjectArrayProperty(
          OptionTags.USER_DEFINED_EDITOR_VARIABLES, UserEditorVariablePO.class);
    for (int i = 0; i < userEditorVars.length; i++) {
      String varName = userEditorVars[i].getVarName();
      expr = Formatter.formatMessage(expr, varName, userEditorVars[i].getVarValue());
    }

    //Expand functions next so that if they contain editor variables, they will also be expanded.
    //${env(VAR_NAME)}
    expr = replaceFunctions(
        expr,
        ENV_FUNCTION_VARIABLE_PREFIX,
        FUNCTION_VARIABLE_SUFFIX,
        new FunctionResolver() {
          @Override
          public String resolveFunctionWithParameter(String param) {
            return System.getenv(param);
          }
        });
    //${system(var.name)}
    expr = replaceFunctions(
        expr,
        SYSTEM_FUNCTION_VARIABLE_PREFIX,
        FUNCTION_VARIABLE_SUFFIX,
        new FunctionResolver() {
          @Override
          public String resolveFunctionWithParameter(String param) {
            return System.getProperty(param);
          }
        });
    
    // ${date(format)}
    expr = replaceFunctions(
        expr,
        DATE_FUNCTION_VARIABLE_PREFIX,
        FUNCTION_VARIABLE_SUFFIX,
        new FunctionResolver() {
          @Override
          public String resolveFunctionWithParameter(String param) {
            try {
              return new SimpleDateFormat(param).format(new Date());
            } catch(Exception ex) {
              logger.warn(ex, ex);
              //Could not resolve
              return "";
            }
          }
        });
    
    // Continue expanding
    
    /*
     * If baseName is not null or empty try to expand CURRENT_FILE_DIRECTORY_URL,
     * CURRENT_FILE_DIRECTORY, CURRENT_FILENAME, CURRENT_FILENAME_WITH_EXTENSION.
     */
    if (currentEditedFileURL != null && currentEditedFileURL.length() > 0) {
      String cfdu = null;
      String cfn = null;
      String cfne = null;
      try {
        String[] pathAndFN = URLUtil.extractPathAndFileName(new URL(currentEditedFileURL));
        cfdu = pathAndFN[0];
        if (cfdu != null && cfdu.endsWith("/")) {
          cfdu = cfdu.substring(0, cfdu.length() - 1);
        }
        cfn = URLUtil.uncorrect(pathAndFN[1]);
        if (cfn != null) {
          cfne = cfn;
          int index1 = cfn.lastIndexOf('.');
          if (index1 != -1) {
            cfn = cfn.substring(0, index1);
          }
        }
      } catch (MalformedURLException e2) {
        logger.warn(e2, e2);
      }
      
      // Extracting the path and filename did not succeeded. Try analyzing the path.
      if (cfdu == null || cfn == null) {
        String path = currentEditedFileURL;
        if (path.startsWith("/")) {
          path = path.substring(1);
        }

        int index = path.lastIndexOf('/');
        if (index != -1) {
          if (cfdu == null) {
            cfdu = path.substring(0, index);
          }
          
          if (cfn == null) {
            cfn = URLUtil.uncorrect(path.substring(index + 1));
            cfne = cfn;  
            int index1 = path.lastIndexOf('.');
            if (index1 != -1 && index1 >= index + 1) {
              cfn = path.substring(index + 1, index1);
            }
          }
        }
      }
  
      //Current file name complete with extension
      String cfd = null;
      try {
        URL cfduURL = null;
        if (cfdu != null) {
          cfdu = URLUtil.correct(cfdu, false);
          cfduURL = new URL(cfdu);
        }
        if (cfduURL != null) {
          if("file".equals(cfduURL.getProtocol())) {
            boolean isDriveLetter = false;
            //EXM-23643 Leave the hanging "/" after the drive letter
            final String cfduURLAsString = cfduURL.toString();
            if (PlatformDetector.isWin32() && cfduURLAsString.endsWith(":")) {
              cfduURL = new URL(cfduURLAsString + "/");
              isDriveLetter = true;
            }

            cfd = URLUtil.getCanonicalFileFromFileUrl(cfduURL).getAbsolutePath();
            if(isDriveLetter && cfd.endsWith("\\") 
                //Remove the hanging "\" if it will be composed with other content.
                && ! expr.endsWith(CURRENT_FILE_DIRECTORY)) {
              //EXM-23643 Remove the hanging "/" after the drive letter
              cfd = cfd.substring(0, cfd.length() - 1);
            }
          } else {
            //EXM-26244 Remote resource, we'll expand ${cfd} to a temporary directory. 
            File tempDir = FileSystemUtil.getOxygenTempDirectory();
            File newFile = new File(tempDir, FileSystemUtil.proposeValidFileName(cfdu));
            cfd = newFile.getAbsolutePath();
          }
        } 
      } catch (MalformedURLException e1) {    
        logger.error("Error when expand " + cfdu + " - " + e1, e1);
      }
  
      expr = slashifyIfNeeded(expr, CURRENT_FILE_DIRECTORY_URL);
      expr = Formatter.formatMessage(expr, CURRENT_FILE_DIRECTORY_URL, cfdu);
      
      if(cfd != null) {
        expr = Formatter.formatMessage(expr, CURRENT_FILE_DIRECTORY, cfd);
      } else {
        //EXM-15566 Leave the editor variable in place.
      }
      
      expr = Formatter.formatMessage(expr, CURRENT_FILENAME, cfn);
      expr = Formatter.formatMessage(expr, CURRENT_FILENAME_WITH_EXTENSION, cfne);
      expr = Formatter.formatMessage(expr, CURRENT_FILE_URL, currentEditedFileURL);
      try {
        //Expand current file.
        File localFile = URLUtil.getCanonicalFileFromFileUrl(new URL(currentEditedFileURL));
        if (localFile != null) {
          expr = Formatter.formatMessage(expr, CURRENT_FILE, localFile.getAbsolutePath());
        }
      } catch (Throwable e) {
        if (logger.isDebugEnabled()) {
          logger.debug("Current file could not be calculated: " + e, e);
        }
      }
      expr = Formatter.formatMessage(expr, CURRENT_FILE_URL_OLD, currentEditedFileURL);
    }

    // Project directory.
    if (expr.indexOf(PROJECT_DIRECTORY) != -1) {
      // Get the current loaded project URL.
      File projectDir = getProjectDirectory(currentEditedFileURL);
      if (projectDir != null) {
        String projectDirectory = projectDir.toString();

        if (projectDirectory.endsWith("/")) {
          projectDirectory = projectDirectory.substring(0, projectDirectory.length() - 1);
        }
        expr = Formatter.formatMessage(expr, PROJECT_DIRECTORY, projectDirectory);
      }
    }

    // Project name.
    if (expr.indexOf(PROJECT_NAME) != -1) {
      String projectName = getProjectName(currentEditedFileURL);
      if (projectName != null) {
        expr = Formatter.formatMessage(expr, PROJECT_NAME, projectName);
      }
    }

    // Project directory URL
    if (expr.indexOf(PROJECT_DIRECTORY_URL) != -1) {
      // Get the current loaded project URL.
      URL projectURL = getCurrentProjectURL(currentEditedFileURL);
      if (projectURL != null) {        
        String projectURLStr = projectURL.toString();
        int index = projectURLStr.lastIndexOf("/");
        if (index != -1) {
          projectURLStr = projectURLStr.substring(0, index);
          // Project directory URL
          expr = Formatter.formatMessage(expr, PROJECT_DIRECTORY_URL, projectURLStr);
        }
      }
    }
    
    // Archive name.
    if (expr.indexOf(ARCHIVE_NAME) != -1) {
      URL currentArchiveURL = getCurrentArchiveURL(currentEditedFileURL);
      if (currentArchiveURL != null) {
        String currentArchiveName = URLUtil.extractFileName(currentArchiveURL);
        if (currentArchiveName != null) {
          currentArchiveName = URLUtil.uncorrect(currentArchiveName);
          int dotIndex = currentArchiveName.lastIndexOf(".");
          if (dotIndex != -1) {
            currentArchiveName = currentArchiveName.substring(0, dotIndex);
          }
          expr = Formatter.formatMessage(expr, ARCHIVE_NAME, currentArchiveName);
        }
      }
    }
    
    // Archive name with extension.
    if (expr.indexOf(ARCHIVE_NAME_WITH_EXTENSION) != -1) {
      URL currentArchiveURL = getCurrentArchiveURL(currentEditedFileURL);
      if (currentArchiveURL != null) {
        String currentArchiveName = URLUtil.extractFileName(currentArchiveURL);
        if (currentArchiveName != null) {
          currentArchiveName = URLUtil.uncorrect(currentArchiveName);
          expr = Formatter.formatMessage(expr, ARCHIVE_NAME_WITH_EXTENSION, currentArchiveName);
        }
      }
    }
    
    // Archive path.
    if (expr.indexOf(ARCHIVE_PATH) != -1) {
      URL currentArchiveURL = getCurrentArchiveURL(currentEditedFileURL);
      if (currentArchiveURL != null) {
        File currentArchivePath = URLUtil.getCanonicalFileFromFileUrl(currentArchiveURL);
        if (currentArchivePath != null) {
          expr = Formatter.formatMessage(expr, ARCHIVE_PATH, currentArchivePath.getAbsolutePath());
        }
      }
    }
    
    // Archive path URL.
    if (expr.indexOf(ARCHIVE_PATH_URL) != -1) {
      URL currentArchiveURL = getCurrentArchiveURL(currentEditedFileURL);
      if (currentArchiveURL != null) {
          expr = Formatter.formatMessage(expr, ARCHIVE_PATH_URL, currentArchiveURL.toString());
      }
    }
    
    // Archive archive file directory.
    if (expr.indexOf(ARCHIVE_FILE_DIRECTORY) != -1) {
      URL currentArchiveURL = getCurrentArchiveURL(currentEditedFileURL);
      if (currentArchiveURL != null) {
        File currentArchivePath = URLUtil.getCanonicalFileFromFileUrl(currentArchiveURL);
        if (currentArchivePath != null) {
          expr = Formatter.formatMessage(
              expr, ARCHIVE_FILE_DIRECTORY, currentArchivePath.getParentFile().getAbsolutePath());
        }
      }
    }
    
    // Archive archive file directory URL.
    if (expr.indexOf(ARCHIVE_FILE_DIRECTORY_URL) != -1) {
      URL currentArchiveURL = getCurrentArchiveURL(currentEditedFileURL);
      if (currentArchiveURL != null) {
        File currentArchivePath = URLUtil.getCanonicalFileFromFileUrl(currentArchiveURL);
        if (currentArchivePath != null) {
          try {
            expr = Formatter.formatMessage(
                expr,
                ARCHIVE_FILE_DIRECTORY_URL,
                URLUtil.correct(currentArchivePath.getParentFile()).toString());
          } catch (MalformedURLException e) {
            logger.error(e, e);
          }
        }
      }
    }
    
    try {
      if (expr.indexOf(USER_HOME_URL) != -1) {
        expr = Formatter.formatMessage(
            expr, USER_HOME_URL, new File(System.getProperty("user.home")).toURL().toString());
      }
    } catch (MalformedURLException e) {
      logger.error("Cannot convert user.home to URL." + e,e);
    }
    
    expr = Formatter.formatMessage(
        expr, USER_HOME_DIR, new File(System.getProperty("user.home")).toString());
    
    String url = System.getProperty(OxygenSystemProperties.COM_OXYGENXML_EDITOR_HOME_URL);
    try {
      URL frameworksURL = getFrameworksUrl();
      
      expr = slashifyIfNeeded(expr, FRAMEWORKS_DIRECTORY_URL);
      expr = Formatter.formatMessage(expr, FRAMEWORKS_DIRECTORY_URL, frameworksURL.toString());
      
      expr = slashifyIfNeeded(expr, FRAMEWORKS_DIRECTORY);
      File frameworksDir = getFrameworksDir();
      if (frameworksDir != null) {
        expr = Formatter.formatMessage(expr, FRAMEWORKS_DIRECTORY, frameworksDir.toString());
      }
    } catch (MalformedURLException e) {
      if (logger.isDebugEnabled()) {
        logger.debug(
            "Cannot convert " + 
            FRAMEWORKS_DIRECTORY_URL +
            " or " + 
            FRAMEWORKS_DIRECTORY +
            ". Home url is " + url);
      }
    }
    
    // oxygenInstallDir
    if (expr.indexOf(OXYGEN_INSTALL_DIR) != -1) {
      // The default is the current directory.
      String homeDir = OxygenDirectoryUtil.getOxygenInstallDir();
      
      expr = Formatter.formatMessage(expr, OXYGEN_INSTALL_DIR, homeDir);
    }
    
    // oxygenHome
    if (expr.indexOf(OXYGEN_HOME_URL) != -1) {
      expr = Formatter.formatMessage(
          expr, OXYGEN_HOME_URL, System.getProperty(OxygenSystemProperties.COM_OXYGENXML_EDITOR_HOME_URL));
    }
    
    //Path separator 
    if (expr.indexOf(PATH_SEPARATOR) != -1) {
      expr = Formatter.formatMessage(expr, PATH_SEPARATOR, getSystemPathSeparator());
    }
    
    // Time Stamp
    if (expr.indexOf(TIMESTAMP) != -1) {
      expr = Formatter.formatMessage(expr, TIMESTAMP, getTimeStamp());
    }
    
    // UUID
    if (expr.indexOf(UUID) != -1) {
      expr = Formatter.formatMessage(expr, UUID, new GenerateEditorVariableValue() {
        
        @Override
        public String generateValue() {
          return java.util.UUID.randomUUID().toString();
        }
      });
    }
    
    // ID custom generated by Oxygen.
    if (expr.indexOf(ID) != -1) {
      expr = Formatter.formatMessage(expr, ID, new GenerateEditorVariableValue() {
        
        @Override
        public String generateValue() {
          return generateUniqueID();
        }
      });
    }
    
    return expr;
  }

  /**
   * Resolves a function
   */
  public static interface FunctionResolver{
    /**
     * Resolve the parameter of a function.
     * @param param The parameter
     * @return The resolved parameter value.
     */
    String resolveFunctionWithParameter(String param);
  }

  /**
   * Replace all functions with the value taken from the map corresponding to the key
   * 
   * @param expr                    The expression to modify.
   * @param functionVariablePrefix  Prefix of function signature.
   * @param functionVariableSuffix  Suffix of function signature.
   * @param keyToValueMap           Map between keys and values.
   *
   * @return The processed expression.
   */
  public static String replaceFunctions(
      String expr,
      String functionVariablePrefix,
      String functionVariableSuffix,
      FunctionResolver keyToValueMap) {
    if (expr.contains(functionVariablePrefix) && expr.contains(functionVariableSuffix)) {
      //Interesting expression
      int index = 0;
      StringBuilder expressionBuffer = new StringBuilder(expr);
      while (index >= 0) {
        index = expressionBuffer.indexOf(functionVariablePrefix, index);
        if (index != -1) {
          //End of function
          int endIndex = expressionBuffer.indexOf(functionVariableSuffix, index);
          if (endIndex != -1) {
            String varName = expressionBuffer.substring(index + functionVariablePrefix.length(), endIndex);
            if (varName.length() > 0) {
              Object valueObj = keyToValueMap.resolveFunctionWithParameter(varName);
              if (valueObj != null) {
                int toReplaceLength = endIndex + functionVariableSuffix.length() - index;
                String val = String.valueOf(valueObj);
                //Replace the function with the value
                expressionBuffer.replace(index, index + toReplaceLength, val);
                //Jump over the expanded value.
                index += val.length();
              } else {
                //Null value, do not expand
                index = endIndex + functionVariableSuffix.length();
              }
            } else {
              //No value specified, do not expand
              index = endIndex + functionVariableSuffix.length();
            }
          } else {
            //Probably malformed.
            break;
          }
        }
      }
      expr = expressionBuffer.toString();
    }
    return expr;
  }

  /**
   * Get the project name.
   * 
   * @param currentEditedFileURL The current edited file URL.
   *
   * @return The project name (without extension). 
   */
  private static String getProjectName(String currentEditedFileURL) {
    String projectName = null;
    URL projectURL = getCurrentProjectURL(currentEditedFileURL);
    if (PlatformDetector.isEclipse()){
      // Is eclipse plugin - use project directory.
      File projectDir = URLUtil.getCanonicalFileFromFileUrl(projectURL);
      if (projectDir != null) {
        if (!projectDir.isDirectory()) {
          projectDir = projectDir.getParentFile();
        }
        projectName = projectDir.getName();
      }
    } else {
      // Get the current loaded project URL.
      if (projectURL != null) {    
        projectName = URLUtil.extractFileName(URLUtil.uncorrect(projectURL.toString()));
        if (projectName != null) {
          int index = projectName.lastIndexOf('.');
          if (index != -1) {
            projectName = projectName.substring(0, index);
          }
        }
      }
    }
    return projectName;
  }
  
  /**
   * Get the project directory file.
   * 
   * @param currentEditedFileURL The URL of the current edited file.
   *
   * @return The {@link File} representing the project directory. 
   */
  private static File getProjectDirectory(String currentEditedFileURL) {
    // Get the current loaded project URL.
    URL projectURL = getCurrentProjectURL(currentEditedFileURL);
    if (projectURL != null) {
      try {
        File projectDir = URLUtil.getCanonicalFileFromFileUrl(projectURL);
        if (projectDir != null) {
          if (!projectDir.isDirectory()) {
            projectDir = projectDir.getParentFile();
          }
        }
        return projectDir;
      } catch (Throwable t) {
        if (logger.isDebugEnabled()) {
          logger.debug("Project dir could not be calculated: " + t, t);
        }
      }
    }
    return null;
  }
  
  /**
   * Make an URL relative to the <code>"frameworks"</code> directory.
   * 
   * @param url The original URL.
   *
   * @return The relative URL to the <code>"frameworks"</code> if possible, otherwise the original URL.
   */
  public static String makeURLRelative2Frameworks(String url) {
    if (url != null) {
      try {
        String frameworksUrl = EditorVariables.getFrameworksUrl().toString();
        // We want to make sure that we don't make a mistake in the next case:
        // - url:        file:/D:/projects/eXml/frameworks2/docbook/xsl/html/docbook_custom.xsl
        // - frameworks: file:/D:/projects/eXml/frameworks
        StringBuilder urlWithEditorVariables = new StringBuilder();
        if (url.startsWith(frameworksUrl) &&
            (url.length() == frameworksUrl.length()
              || frameworksUrl.endsWith("/")
              || url.charAt(frameworksUrl.length()) == '/'
              || url.charAt(frameworksUrl.length()) == '\\')) {
          urlWithEditorVariables.append(EditorVariables.FRAMEWORKS_DIRECTORY_URL);
          if (frameworksUrl.endsWith("/")) {
            urlWithEditorVariables.append("/");
          }
          urlWithEditorVariables.append(url.substring(frameworksUrl.length()));
          url = urlWithEditorVariables.toString();
        }
      } catch (MalformedURLException e) {
        // Leave it as it was.
      }
    }
    return url;
  }
  
  /**
   * Make a file or directory relative to the <code>"frameworks"</code> directory.
   * 
   * @param fileOrDir The original file or directory.
   *
   * @return The relative path to the <code>"frameworks"</code> if possible, otherwise the original
   *         file or directory path.
   */
  public static String makeFileRelative2Frameworks(String fileOrDir) {
    if (fileOrDir != null) {
      try {
        //First make both files canonical.
        String frameworksDir = EditorVariables.getFrameworksDir().getCanonicalPath();
        String fileOrDirCan = new File(fileOrDir).getCanonicalPath();
        StringBuilder urlWithEditorVariables = new StringBuilder();
        if (fileOrDirCan != null 
            && fileOrDirCan.startsWith(frameworksDir)
            && (fileOrDirCan.length() == frameworksDir.length() 
                || fileOrDirCan.charAt(frameworksDir.length()) == '/'
                || fileOrDirCan.charAt(frameworksDir.length()) == '\\')) {
          urlWithEditorVariables.append(EditorVariables.FRAMEWORKS_DIRECTORY);
          urlWithEditorVariables.append(fileOrDir.substring(frameworksDir.length()));
          fileOrDir = urlWithEditorVariables.toString();
          //EXM-11773 This path saved on Windows should work also on Linux
          fileOrDir = Util.replaceAll(fileOrDir, "\\", "/");
        }
      } catch (IOException e) {
        // Leave it as it was.
      }
    }
    return fileOrDir;
  }
  
  /**
   * Used from tests so we may set a dir independently of the home.
   */
  private static URL frameworksURLForTests = null;
  
  /**
   * Set a frameworks url so it will not be computed from the home url.
   * 
   * @param fURL The url.
   */
  public static void setFrameworksURLForTest(URL fURL) {
    frameworksURLForTests = fURL;
  }
  
  /**
   * Get the <code>"frameworks"</code> directory {@link URL}.
   * 
   * @return The <code>"frameworks"</code> directory {@link URL}.
   *
   * @throws MalformedURLException If the oxygen home URL is not set.
   */
  public static URL getFrameworksUrl() throws MalformedURLException {
    if (frameworksURLForTests != null) {
      return frameworksURLForTests;
    }
    return getCurrentFrameworksURL();
  }
  
  /**
   * <code>true</code> if the frameworks directory is initializing
   */
  private static boolean frameworksDirInitializing = false;

  /**
   * Get the current frameworks directory.
   *
   * @return The current frameworks URL. Can be a custom one if it was set in options, or the default one.
   *
   * @throws MalformedURLException If the oxygen home URL is not set.
   */
  public static URL getCurrentFrameworksURL() throws MalformedURLException {
    URL toReturn =  null;
    if (!frameworksDirInitializing) {
      if(! frameworksDirInitialized) {
        if(System.getenv(OxygenSystemProperties.COM_OXYGENXML_EDITOR_CUSTOM_FRAMEWORKS_URL) != null) {
          //Set from outside as an environment variable.
          //Set it also as a system property.
          System.setProperty(
              OxygenSystemProperties.COM_OXYGENXML_EDITOR_CUSTOM_FRAMEWORKS_URL, 
              System.getenv(OxygenSystemProperties.COM_OXYGENXML_EDITOR_CUSTOM_FRAMEWORKS_URL));
        }
        //First time initialization 
        if (
            //Maybe the System property was set from outside, it should be stronger than the user options.
            System.getProperty(
                OxygenSystemProperties.COM_OXYGENXML_EDITOR_CUSTOM_FRAMEWORKS_URL) == null
                //Custom user frameworks.
            && Options.getInstance().getBooleanProperty(OptionTags.USE_CUSTOM_FRAMEWORKS_DIRECTORY_URL)) {
          //Try the version with editor variables.
          String withEditorVariables = Options.getInstance().getStringProperty(
              OptionTags.FRAMEWORKS_DIRECTORY_URL_WITH_EDITOR_VARIABLES);
          if (withEditorVariables != null) {
            //Expand the editor variables
            try {
              frameworksDirInitializing = true;
              //System.err.println("WITH VARS BEFORE " + withEditorVariables);
              withEditorVariables = expandEditorVariables(withEditorVariables, null);
              //System.err.println("WITH VARS AFTER " + withEditorVariables);
            } finally {
              frameworksDirInitializing = false;
            }
            try {
              toReturn = new URL(withEditorVariables);
            } catch(MalformedURLException ex) {
              logger.error(ex, ex);
              try {
                toReturn = URLUtil.correct(new File(withEditorVariables));
              } catch(MalformedURLException e) {
                logger.error(e, e);
              }
            }
          }
          // Use custom frameworks directory.
          if (toReturn == null) {
            //Use deprecated way of storing
            toReturn = Options.getInstance().getURLProperty(OptionTags.FRAMEWORKS_DIRECTORY_URL);
          }
        }

        // Set it in the system properties to be used until next restart.
        if (toReturn != null) {
          System.setProperty(
              OxygenSystemProperties.COM_OXYGENXML_EDITOR_CUSTOM_FRAMEWORKS_URL, toReturn.toString());
        }
        frameworksDirInitialized = true;
      }

      // If was initialized get it from the system properties.
      String url = System.getProperty(
          OxygenSystemProperties.COM_OXYGENXML_EDITOR_CUSTOM_FRAMEWORKS_URL);
      try {
        toReturn = new URL(URLUtil.correct(url, false));
      } catch (MalformedURLException e) {
        // Use the default one.
      }
    }
    // Use the default one.
    if (toReturn == null) {
      toReturn = getDefaultFrameworkURL();
    }
    return toReturn;
  }

  /**
   * Get the default frameworks directory. 
   * It is relative to the oxygen installation directory.
   * 
   * @return The default frameworks directory.
   *
   * @throws MalformedURLException If the oxygen home URL is not set.
   */
  private static URL getDefaultFrameworkURL() throws MalformedURLException {
    String url = System.getProperty(OxygenSystemProperties.COM_OXYGENXML_EDITOR_HOME_URL);
    return new URL(new URL(URLUtil.correct(url, false)), "frameworks");
  }
  
  /**
   * Used from tests so we may set a dir independently of the home.
   */
  private static File frameworksDirForTests = null;
  
  /**
   * Set a frameworks dir so it will not be computed from the home url.
   * 
   * @param fDir
   */
  public static void setFrameworksDirForTest(File fDir) {
    frameworksDirForTests = fDir;
  }
  
  /**
   * Get the {@link File} representing the <code>"frameworks"</code> directory.
   * 
   * @return The <code>"frameworks"</code> directory {@link File}.
   *
   * @throws MalformedURLException If the oxygen home URL is not set.
   */
  public static File getFrameworksDir() throws MalformedURLException {
    if (frameworksDirForTests != null) {
      return frameworksDirForTests;
    }
    return URLUtil.getCanonicalFileFromFileUrl(getCurrentFrameworksURL());
  }
  
  /**
   * Get the base frameworks directory from the user preferences directory like:
   * Users\***\AppData\Roaming\com.oxygenxml\extensions\v14.0\frameworks
   * 
   * Note: This is not the actual directory with frameworks but the directory which contains
   * all frameworks directories. A directory for each update site:
   * - extensions\v14.0\frameworks\{update_site_1}
   * - extensions\v14.0\frameworks\{update_site_2}
   * 
   * @return The user frameworks directory.
   */
  public static File getBaseUserFrameworksDir() {
    return new File(PreferencesDirectoryDetector.getInstance().getCurrentExtensionsDirectory(), "frameworks");
  }
  
  /**
   * Get the frameworks directories from the user preferences directory like:
   * Users\***\AppData\Roaming\com.oxygenxml\extensions\v14.0\frameworks\{update_site}.
   * A new level is inserted ({update_site}) to protect against add-ons conflicts (sort of like a namespace).  
   * 
   * @return The user frameworks directories or null if it doesn't exists.
   */
  public static File[] getUserFrameworksDirs() {
    File abstractFrameworks = getBaseUserFrameworksDir();
    return abstractFrameworks.listFiles(new ExceptCVSAndSVNFileFilter() {
      @Override
      public boolean accept(File pathname) {
        return pathname.isDirectory() && super.accept(pathname);
      }
    });
  }
  
  /**
   * Returns the {@link URL} of the current project.
   * 
   * @param currentFileSystemID  The current file system ID.
   *
   * @return The current project {@link URL} or <code>null</code> if it cannot be determined.
   */
  public static URL getCurrentProjectURL(String currentFileSystemID) {
    URL currentProjectURL = null;
    if (projectURLProvider != null) {
      currentProjectURL = projectURLProvider.getProjectURL(currentFileSystemID);
    } else {
      //Either we are in EC with no project URL support      
      if (!PlatformDetector.isEclipse()) {
        //Or we are on SA before the provider was set (like when expand macros is called
        //from EXMLResourceBundle
        currentProjectURL = Options.getInstance().getURLProperty(OptionTags.LAST_PROJECT);
      }
    }
    if (currentProjectURL == null || "file:newProject.xpr".equals(currentProjectURL.toString())) {
      try {
        return new File(System.getProperty("user.dir", ".")).toURL();
      } catch (MalformedURLException e) {
        // Neglect.
      }
    }
    return currentProjectURL;
  }
  
  /**
   * Returns the {@link URL} of the current archive.
   * 
   * @param currentFileSystemID  The current file system ID.
   *
   * @return The current archive {@link URL} or <code>null</code> if it cannot be determined.
   */
  public static URL getCurrentArchiveURL(String currentFileSystemID) {
    URL currentArchiveURL = null;
    if (currentFileSystemID != null) {
      if(currentFileSystemID.startsWith("zip:")) {
        //The current opened file is from an archive, find out the archive URL.
        int archiveEndIndex = currentFileSystemID.indexOf("!/");
        try {
          if (archiveEndIndex != -1) {
            currentArchiveURL =
              new URL(currentFileSystemID.substring("zip:".length(), archiveEndIndex));
          } else {
            //Maybe it is an archive by itself.
            currentArchiveURL =
              new URL(currentFileSystemID.substring("zip:".length(), currentFileSystemID.length()));
          }
        } catch (MalformedURLException e) {
          logger.error(e, e);
        }
      } else {
        //Maybe file path to archive.
        if(archiveExtensionsRecognizer != null 
            && archiveExtensionsRecognizer.hasArchiveExtension(currentFileSystemID)) {
          //It is itself an archive
          try {
              currentArchiveURL = new URL(currentFileSystemID);
          } catch (MalformedURLException e) {
            logger.error(e, e);
          }
        }
      }
    }
    
    //Fall back to the current URL opened in the archive browser view.
    if (currentArchiveURL == null && archiveURLProvider != null) {
      currentArchiveURL = archiveURLProvider.getArchiveURL();
    }
    return currentArchiveURL;
  }

  /**
   * Append end slash if the <code>expr</code> ends with the <code>editorVariable</code>
   * 
   * @param expr            The expression to analyze.
   * @param editorVariable  The editor variable to be expanded.
   *
   * @return The expression or the expression with '/' character appended.
   */
  private static String slashifyIfNeeded(String expr, String editorVariable) {
    if (expr.endsWith(editorVariable)) {
      expr += "/";
    }
    return expr;
  }

  /**
   * Set the project URL provider.
   * 
   * @param projectURLProvider The new project URL provider.
   */
  public static void setProjectURLProvider(ProjectURLProvider projectURLProvider) {
    EditorVariables.projectURLProvider = projectURLProvider;
  }
  
  /**
   * Set the archive URL provider.
   * 
   * @param archiveURLProvider The new archive URL provider.
   */
  public static void setArchiveURLProvider(ArchiveURLProvider archiveURLProvider) {
    EditorVariables.archiveURLProvider = archiveURLProvider;
  }
  
  /**
   * Set an archive extensions recognizer
   * 
   * @param archiveExtensionsRecognizer The archiveExtensionsRecognizer.
   */
  public static void setArchiveExtensionsRecognizer(
      ArchiveExtensionsRecognizer archiveExtensionsRecognizer) {
    EditorVariables.archiveExtensionsRecognizer = archiveExtensionsRecognizer;
  }
  
  /**
   * @return The system path separator. It is dependent on the platform.
   */
  public static String getSystemPathSeparator() {
    return File.pathSeparator;
  }
  
  /**
   * @return The current date and time formatted as String. 
   * The used format is <code>yyyyMMddHHmmssSSS</code>.
   */
  static String getTimeStamp() { 
   SimpleDateFormat date = new SimpleDateFormat("yyyyMMddHHmmssSSS");
   return date.format(Calendar.getInstance().getTime());
  }
  

  //The number of different values for a rank
  private final static int BASE = 'z'-'a' + 1;
  //Subtracted from the current timestamp to obtain smaller IDs
  private final static long OFFSET = 1286892000000L;
  //Used to avoid duplicate IDs
  private static long lastUsedTimestamp = 0;
  
  /**
   * @return The generated Unique ID
   */
  public static synchronized String generateUniqueID() {
    String uid = "";
    long time = System.currentTimeMillis() - OFFSET;
    if (time <= lastUsedTimestamp) {
      time = lastUsedTimestamp + 1;
    }
    int underscores = 0;
    lastUsedTimestamp = time;
    while (time != 0) {
      int mod = (int) (time % BASE);
      char ch = (char) ('a' + mod);
      // EXM-24105 Check if it is the first letter from id and do not transform it from vowel to int
      if (uid.isEmpty()) {
        uid = uid + ch;
      } else {
        uid = uid + convertVowelToInt(ch);
      }
      time = time / BASE;
      if (time > 0 && (uid.length() - underscores)% 3 == 0){
        uid += "_";
        underscores ++;
      }
    }
    return uid;
  }

  /**
   * Convert a vowel into an integer value.
   * 
   * @param ch The character to be converted.
   *
   * @return The corresponding integer if <code>ch</code> is a vowel, otherwise <code>ch</code> itself.  
   */
  private static char convertVowelToInt(char ch) {
    char toReturn = ch;
    switch (ch) {
      case 'a':
        toReturn = '1';
        break;
      case 'e':
        toReturn = '2';
        break;
      case 'i':
        toReturn = '3';
        break;
      case 'o':
        toReturn = '4';
        break;
      case 'u':
        toReturn = '5';
        break;
    }
    return toReturn;
  }
  
  /**
   * Expand FRAMEWORKS_DIRECTORY_URL, FRAMEWORKS_DIRECTORY, FRAMEWORK_DIRECTORY and FRAMEWORK_URL.
   * All these variables are resolved relative to framework store location.
   * 
   * @param expr Editor variables expression.
   * @param frameworkStoreLocation The framework store location as an file path. 
   * @param rewritePolicy Rewrite FRAMEWORK_DIRECTORY and FRAMEWORK_URL 
   * using FRAMEWORKS_DIRECTORY_URL and FRAMEWORKS_DIRECTORY variables.
   * 
   * @return The given expression with the framework variables expanded.
   */
  public static String expandFrameworksVariables(
      String expr, 
      String frameworkStoreLocation,
      FrameworkRewritePolicy rewritePolicy) {
    if (frameworkStoreLocation != null) {
      String url = System.getProperty(OxygenSystemProperties.COM_OXYGENXML_EDITOR_HOME_URL);
      try {
        File frameworkDir = null;
        URL frameworkURL = null;
        try {
          URL frameworkStoreLocationURL = new URL(frameworkStoreLocation);
          frameworkDir = URLUtil.getCanonicalFileFromFileUrl(frameworkStoreLocationURL).getParentFile();
        } catch (MalformedURLException e) {
          // Must be a file path
          frameworkDir = new File(frameworkStoreLocation).getParentFile();
        }
        frameworkURL = URLUtil.correct(frameworkDir);
        File frameworksDir = frameworkDir.getParentFile();
        URL frameworksURL = URLUtil.correct(frameworksDir);

        return expandFrameworksVariables(expr, 
            rewritePolicy, 
            frameworksDir, 
            frameworksURL, 
            frameworkDir, 
            frameworkURL);

      } catch (MalformedURLException e) {
        if (logger.isDebugEnabled()) {
          logger.debug("Cannot convert " + FRAMEWORKS_DIRECTORY_URL + " or " + FRAMEWORKS_DIRECTORY
              + ". Home url is " + url);
        }
      }
    }
    return expr;
  }
  
  /**
   * Expand FRAMEWORKS_DIRECTORY_URL, FRAMEWORKS_DIRECTORY, FRAMEWORK_DIRECTORY and FRAMEWORK_URL.
   * 
   * @param expr Editor variables expression.
   * @param rewritePolicy Rewrite FRAMEWORK_DIRECTORY and FRAMEWORK_URL 
   * using FRAMEWORKS_DIRECTORY_URL and FRAMEWORKS_DIRECTORY variables.
   * @param frameworksDir The directory where all frameworks reside.
   * @param frameworksURL The URL of the directory where all frameworks reside.
   * @param frameworkDir The directory where the specific framework reside.
   * @param frameworkURL The URL of the directory where the specific framework reside.
   * 
   * @return The given expression with the framework variables expanded.
   */
  public static String expandFrameworksVariables(
      String expr, 
      FrameworkRewritePolicy rewritePolicy, 
      File frameworksDir,
      URL frameworksURL,
      File frameworkDir,
      URL frameworkURL) {
    if (frameworkURL != null && expr != null) {
      String url = System.getProperty(OxygenSystemProperties.COM_OXYGENXML_EDITOR_HOME_URL);
      try {
        String stringRepres = frameworkURL.toExternalForm();
        if (stringRepres.endsWith("/")) {
          // File.toURL will add a terminating /
          frameworkURL = new URL(stringRepres.substring(0, stringRepres.length() - 1));
        }

        stringRepres = frameworksURL.toExternalForm();
        if (stringRepres.endsWith("/")) {
          // File.toURL will add a terminating /
          frameworksURL = new URL(stringRepres.substring(0, stringRepres.length() - 1));
        }
        
        String frameworkDirAsString = frameworkDir.toString().replace('\\', '/');
        if (rewritePolicy != FrameworkRewritePolicy.REWRITE_WITH_FRAMEWORK) {
          // If we are allowed to expand the $framework variables.
          String frameworkURLReplace = frameworkURL.toString();
          if (rewritePolicy == FrameworkRewritePolicy.REWRITE_WITH_FRAMEWORKS) {
            frameworkURLReplace = makeURLRelative2Frameworks(frameworkURLReplace);
          }

          expr = slashifyIfNeeded(expr, FRAMEWORK_URL);
          expr = Formatter.formatMessage(expr, FRAMEWORK_URL, frameworkURLReplace);

          expr = slashifyIfNeeded(expr, FRAMEWORK_DIRECTORY);
          if (frameworkDir != null) {
            String frameworkDirReplace = frameworkDirAsString;
            if (rewritePolicy == FrameworkRewritePolicy.REWRITE_WITH_FRAMEWORKS) {
              frameworkDirReplace = makeFileRelative2Frameworks(frameworkDirReplace);
            }
            expr = Formatter.formatMessage(expr, FRAMEWORK_DIRECTORY, frameworkDirReplace);
          }
        }

        if (rewritePolicy != FrameworkRewritePolicy.REWRITE_WITH_FRAMEWORKS) {
          // Replace frameworks variables.
          expr = slashifyIfNeeded(expr, FRAMEWORKS_DIRECTORY_URL);
          String originalExpr = expr;
          expr = Formatter.formatMessage(expr, FRAMEWORKS_DIRECTORY_URL, frameworksURL.toString());
          String expandedExpr = expr;
          if (rewritePolicy != FrameworkRewritePolicy.REWRITE_ABSOLUTE) {
            expr = Formatter.formatMessage(expr, frameworkURL.toExternalForm() + "/", FRAMEWORK_URL + "/");
            if (expr.equals(expandedExpr)) {
              // If the expression could not be compressed to ${framework}, leave it as it was.
              // This is the case for referring resources from another framework through the use of ${frameworks} 
              if (expandedExpr.equals(frameworkURL.toExternalForm())) {
                // For example we try to map ${frameworks}/docbook with ${framework} 
                // (and there is no trailing "/" after "docbook"). The leading slash
                // is required for a situation like this:
                // ${framework} = /home/frameworks/doc then ${frameworks}/docbook must 
                // not match ${framework}.
                expr = FRAMEWORK_URL + "/";
              } else {
                // If the expression could not be compressed to ${frameworkDir}, leave it as it was.
                // This is the case for referring resources from another framework through the use of ${frameworksDir} 
                expr = originalExpr;
              }
            }
          }

          expr = slashifyIfNeeded(expr, FRAMEWORKS_DIRECTORY);
          if (frameworksDir != null) {
            String frameworksDirAsString = frameworksDir.toString().replace('\\', '/');
            originalExpr = expr;
            expr = Formatter.formatMessage(expr, FRAMEWORKS_DIRECTORY, frameworksDirAsString);
            expandedExpr = expr;
            if (rewritePolicy != FrameworkRewritePolicy.REWRITE_ABSOLUTE) {
              expr = Formatter.formatMessage(expr, frameworkDirAsString + "/", FRAMEWORK_DIRECTORY + "/");
              if (expr.equals(expandedExpr)) {
                if (expandedExpr.equals(frameworkDirAsString)) {
                  // For example we try to map ${frameworks}/docbook with ${framework} 
                  // (and there is no trailing "/" after "docbook"). The leading slash
                  // is required for a situation like this:
                  // ${framework} = /home/frameworks/doc then ${frameworks}/docbook must 
                  // not match ${framework}.
                  expr = FRAMEWORK_DIRECTORY + "/";
                } else {
                  // If the expression could not be compressed to ${frameworkDir}, leave it as it was.
                  // This is the case for referring resources from another framework through the use of ${frameworksDir} 
                  expr = originalExpr;
                }
              }
            }
          }
        } 
      } catch (MalformedURLException e) {
        if (logger.isDebugEnabled()) {
          logger.debug("Cannot convert " + FRAMEWORKS_DIRECTORY_URL + " or " + FRAMEWORKS_DIRECTORY
              + ". Home url is " + url);
        }
      }
    }
    return expr;
  }

  /**
   * Make an URL relative to $framework or that if not possible, to $frameworks (also if possible).
   * OBS: $frameworks is the frameworks directory of the given framework frameworkFileURL.
   * 
   * @param url The URL.
   * @param frameworkFileURL The URL of the ".framework" file.
   * 
   * @return The relative URL to the $framework or $frameworks if possible, otherwise the original URL.
   */
  public static String makeURLRelative2Framework(String url, String frameworkFileURL) {
    if (url != null && frameworkFileURL != null) {
      StringBuilder urlWithEditorVariables = new StringBuilder();
      String frameworkDirURL = frameworkFileURL.substring(0, frameworkFileURL.lastIndexOf("/") + 1);
      if (url.startsWith(frameworkDirURL)) {
        urlWithEditorVariables.append(EditorVariables.FRAMEWORK_URL);
        urlWithEditorVariables.append("/");
        urlWithEditorVariables.append(url.substring(frameworkDirURL.length()));
        url = urlWithEditorVariables.toString();
      } else {
        // Try to make it relative to the frameworks directory of the document type.
        try {
          URL frameworkStoreLocationURL = new URL(frameworkFileURL);
          File frameworkDir = URLUtil.getAbsoluteFileFromFileUrl(frameworkStoreLocationURL).getParentFile();
          File frameworksDir = frameworkDir.getParentFile().getCanonicalFile();
          String frameworksURL = URLUtil.correct(frameworksDir).toExternalForm();

          if (url.startsWith(frameworksURL)
              && (url.length() == frameworksURL.length()
                  || frameworksURL.endsWith("/")
                  || url.charAt(frameworksURL.length()) == '\\'
                  || url.charAt(frameworksURL.length()) == '/')) {
            urlWithEditorVariables.append(EditorVariables.FRAMEWORKS_DIRECTORY_URL);
            if (frameworksURL.endsWith("/")) {
              urlWithEditorVariables.append("/");
            }
            urlWithEditorVariables.append(url.substring(frameworksURL.length()));
            url = urlWithEditorVariables.toString();
          }
        } catch (IOException e) {
          // Shouldn't happen.
          logger.error(e, e);
        }
      }
    }
    
    return url;
  }
  
  /**
   * Make a file or directory relative to the $framework or $frameworks directory.
   * The $frameworks directory is the frameworks directory of the given framework file,
   * so it might differ from the default Oxygen frameworks directory.
   * 
   * @param path The original file or directory.
   * @param frameworkFilePath The URL of the ".framework" file.
   *
   * @return The relative path to the $framework or $frameworks if possible, otherwise the original
   *         file or directory path.
   */
  public static String makeFileRelative2Framework(String path, String frameworkFilePath) {
    if (path != null && frameworkFilePath != null) {
      try {
        //First make both files canonical.
        String frameworkDir = new File(frameworkFilePath).getParentFile().getCanonicalPath();
        String fileOrDirCan = new File(path).getCanonicalPath();
        StringBuilder urlWithEditorVariables = new StringBuilder();
        if (fileOrDirCan != null 
            && fileOrDirCan.startsWith(frameworkDir)
            && (fileOrDirCan.length() == frameworkDir.length() 
                || fileOrDirCan.charAt(frameworkDir.length()) == '/'
                  || fileOrDirCan.charAt(frameworkDir.length()) == '\\')) {
          urlWithEditorVariables.append(EditorVariables.FRAMEWORK_DIRECTORY);
          urlWithEditorVariables.append(path.substring(frameworkDir.length()));
          path = urlWithEditorVariables.toString();
          //EXM-11773 This path saved on Windows should work also on Linux
          path = Util.replaceAll(path, "\\", "/");
        } else {
          // Try to make the path relative to the frameworks directory of the 
          // current DT.
          String frameworksDir = new File(frameworkDir).getParentFile().getCanonicalPath();
          if (fileOrDirCan != null 
              && fileOrDirCan.startsWith(frameworksDir)
              && (fileOrDirCan.length() == frameworksDir.length() 
                  || fileOrDirCan.charAt(frameworksDir.length()) == '/'
                    || fileOrDirCan.charAt(frameworksDir.length()) == '\\')) {
            urlWithEditorVariables.append(EditorVariables.FRAMEWORKS_DIRECTORY);
            urlWithEditorVariables.append(path.substring(frameworksDir.length()));
            path = urlWithEditorVariables.toString();
            //EXM-11773 This path saved on Windows should work also on Linux
            path = Util.replaceAll(path, "\\", "/");
          }
        }
      } catch (IOException e) {
        // Leave it as it was.
      }
    }
    
    return path;
  }
}